r"""
Time series module
"""

__all__= []

from .main import *
from .main import __all__ as _main__all__

from .functions import *
from .functions import __all__ as _functions__all__

__all__.extend(_main__all__)
__all__.extend(_functions__all__)




"""
"""


#[
from __future__ import annotations

from ..sequentials import main as _sequentials
#]


class ProcessorMixin:
    """
    """
    #[

    @property
    def preprocessor(self, /, ) -> _sequentials.Sequential:
        """
        """
        return self._invariant.preprocessor

    @property
    def postprocessor(self, /, ) -> _sequentials.Sequential:
        """
        """
        return self._invariant.postprocessor

    @preprocessor.setter
    def preprocessor(
        self,
        preprocessor: _sequentials.Sequential,
        /,
    ) -> None:
        """
        """
        self._invariant.preprocessor = preprocessor

    @postprocessor.setter
    def postprocessor(
        self,
        postprocessor: _sequentials.Sequential,
        /,
    ) -> None:
        """
        """
        self._invariant.postprocessor = postprocessor

    def preprocess(
        self,
        *args,
        **kwargs,
    ) -> tuple:
        return self.preprocessor.simulate(*args, **kwargs, )

    def postprocess(
        self,
        *args,
        **kwargs,
    ) -> tuple:
        return self.postprocessor.simulate(*args, **kwargs, )

    #]




r"""
Slatable protocol
"""


#[

from __future__ import annotations

#]


class Slatable:
    r"""
    """
    #[

    max_lag: int = 0
    max_lead: int = 0
    databox_names: tuple[str, ...] = ()
    output_names: tuple[str, ...] = ()
    fallbacks: dict[str, Real] | None = None
    overwrites: dict[str, Real] | None = None
    qid_to_logly: dict[str, bool] | None = None
    databox_validators: dict[str, Callable] | None = None
    descriptions: tuple[str, ...] = ()

    #]




r"""
"""


#[

from __future__ import annotations

import documark as _dm

from ..dates import Period, Frequency

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Callable

#]


def mixin(klass: type, ) -> type:
    r"""
    Inlay the steady state methods in the class
    """
    #[
    klass.to_jsonable = to_jsonable
    klass.from_jsonable = classmethod(from_jsonable, )
    return klass
    #]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Series class
#-------------------------------------------------------------------------------


@_dm.reference(category="serialization", )
def to_jsonable(
    self,
    *,
    period_to_string: Callable = Period.to_sdmx_string,
    include_description: bool = True,
    include_frequency: bool = True,
    allow_multiple_variants: bool = False,
) -> dict:
    r"""
................................................................................

==Convert time series to a JSON-serializable dictionary==

This method converts the time series into a JSON-serializable dictionary format.
It includes options to include or exclude the description and frequency of the
series, as well as handling multiple variants.

    jsonable = self.to_jsonable(
        period_to_string=Period.to_sdmx_string,
        include_description=True,
        include_frequency=True,
        allow_multiple_variants=False,
    )


### Input arguments ###

???+ input "period_to_string"
    Function to convert a Period object to a string; choose from:
    * Period.to_sdmx_string (default): Converts to SDMX format (e.g., "2023-Q1").
    * Period.to_iso_string: Converts to ISO format (e.g., "2023-01-01").
    * custom function taking a Period and returning a string.

???+ input "include_description"
    If `True` (default), the `jsonable` dictionary includes the series
    description.

???+ input "include_frequency"
    If `True` (default), the `jsonable` dictionary includes the series
    frequency; this is recommended for clarity and robustness.

???+ input "allow_multiple_variants"
    Not implemented yet.


### Returns ###

???+ returns "jsonable"
    A dictionary representing the time series in a JSON-serializable format.

................................................................................
    """
    #[
    if allow_multiple_variants:
        raise NotImplementedError("allow_multiple_variants=True is not implemented yet.")
    #
    if self.num_variants > 1 and not allow_multiple_variants:
        raise ValueError("The series has multiple variants. Set allow_multiple_variants=True to serialize all variants.")
    #
    jsonable = {}
    #
    if include_description:
        jsonable["description"] = self.get_description() or ""
    #
    if include_frequency:
        jsonable["frequency"] = self.frequency.to_jsonable()
    #
    jsonable["start"] = (
        period_to_string(self.start)
        if self.start is not None else None
    )
    #
    jsonable["values"] = self.get_values(unpack_singleton=not allow_multiple_variants)
    #
    return jsonable
    #]


@_dm.reference(
    category="constructor",
    call_name="Series.from_jsonable",
)
def from_jsonable(
    klass,
    jsonable: dict,
    *,
    period_from_string: Callable = Period.from_sdmx_string,
    frequency_included: bool = True,
    description_included: bool = True,
    allow_multiple_variants: bool = False,
) -> Self:
    r"""
................................................................................

==Create time series from a JSON-serializable dictionary==

This class method creates a time series from a JSON-serializable dictionary.

    self = Series.from_jsonable(
        jsonable,
        period_from_string=Period.from_sdmx_string,
        frequency_included=True,
        description_included=True,
        allow_multiple_variants=False,
    )


### Input arguments ###

???+ input "jsonable"
    A dictionary representing the time series in a JSON-serializable format.

???+ input "period_from_string"
    Function to convert a string to a Period object; choose from:
    * Period.from_sdmx_string (default): Parses SDMX format (e.g., "2023-Q1").
    * Period.from_iso_string: Parses ISO format (e.g., "2023-01-01"), needs a `frequency` to be included for non-daily periods.
    * custom function taking a string and a frequency, and returning a Period.

???+ input "frequency_included"
    If `True` (default), the `jsonable` dictionary is expected to include the
    series frequency represented as a single letter.

???+ input "description_included"
    If `True` (default), the `jsonable` dictionary is expected to include the
    series description.

???+ input "allow_multiple_variants"
    Not implemented yet.


### Returns ###

???+ returns "self"
    A time series object created from the `jsonable` dictionary.

................................................................................
    """
    #[
    if allow_multiple_variants:
        raise NotImplementedError("allow_multiple_variants=True is not implemented yet.")
    #
    frequency = (
        Frequency.from_jsonable(jsonable["frequency"], )
        if frequency_included else None
    )
    #
    start = (
        period_from_string(jsonable["start"], frequency=frequency, )
        if jsonable["start"] else None
    )
    #
    values = jsonable["values"]
    if not allow_multiple_variants:
        values = tuple(values)
    #
    description = (
        jsonable["description"]
        if description_included else None
    )
    #
    if not start or not values:
        return klass.as_empty(description=description, )
    #
    return klass(
        start=start,
        values=values,
        description=description,
    )
    #]


#-------------------------------------------------------------------------------




"""
Model quantities
"""


#[
from __future__ import annotations

from typing import Self, Any
from typing import Protocol
from collections.abc import Iterable, Iterator, Sequence
import enum
import collections as _co
import re as _re
from dataclasses import dataclass

from . import wrongdoings as _wrongdoings
from . import attributes as _attributes
#]


QUANTITY_OCCURRENCE_PATTERN = _re.compile(
    r"\b([a-zA-Z]\w*)\b(\[[-+\d]+\])?(?!\()"
)


_PortableType = tuple[str, str, bool, str, str, ]


class QuantityKind(enum.Flag):
    """
    Classification of model quantities
    """
    #[

    UNSPECIFIED = enum.auto()

    TRANSITION_VARIABLE = enum.auto()
    MEASUREMENT_VARIABLE = enum.auto()
    TRANSITION_SHOCK = enum.auto()
    ANTICIPATED_SHOCK_VALUE = enum.auto()
    MEASUREMENT_SHOCK = enum.auto()
    LHS_VARIABLE = enum.auto()
    RHS_ONLY_VARIABLE = enum.auto()
    PARAMETER = enum.auto()
    EXOGENOUS_VARIABLE = enum.auto()
    TRANSITION_STD = enum.auto()
    MEASUREMENT_STD = enum.auto()

    ENDOGENOUS_VARIABLE = TRANSITION_VARIABLE | MEASUREMENT_VARIABLE
    ANY_VARIABLE = ENDOGENOUS_VARIABLE | EXOGENOUS_VARIABLE
    ANY_SHOCK = TRANSITION_SHOCK | MEASUREMENT_SHOCK
    ANY_SHOCK_OR_SHOCK_VALUE = ANY_SHOCK | ANTICIPATED_SHOCK_VALUE
    ANY_STD = TRANSITION_STD | MEASUREMENT_STD
    PARAMETER_OR_STD = PARAMETER | ANY_STD
    LOGGABLE_VARIABLE = TRANSITION_VARIABLE | MEASUREMENT_VARIABLE | EXOGENOUS_VARIABLE

    @classmethod
    def from_keyword(
        klass,
        keyword: str,
        /,
    ) -> Self:
        """
        """
        return klass[
            keyword
            .replace("-", "_")
            .replace(" ", "_")
            .replace("!", "")
            .strip()
            .upper()
            .removesuffix("S")
        ]

    def to_keyword(self, /, ) -> str:
        return "!" + self.name.lower() + "s"

    def to_portable(self, /, ) -> str:
        return _TO_PORTABLES[self]

    @classmethod
    def from_portable(klass, portable: str, /, ) -> Self:
        return _FROM_PORTABLES[portable]

    @property
    def human(self, /, ) -> str:
        return self.name.replace("_", " ").title()

    #]


for n in QuantityKind.__members__:
    exec(f"{n} = QuantityKind.{n}")


_TO_PORTABLES = {
    QuantityKind.TRANSITION_VARIABLE: "#x",
    QuantityKind.MEASUREMENT_VARIABLE: "#y",
    QuantityKind.TRANSITION_SHOCK: "#u",
    QuantityKind.ANTICIPATED_SHOCK_VALUE: "#v",
    QuantityKind.MEASUREMENT_SHOCK: "#w",
    QuantityKind.PARAMETER: "#p",
    QuantityKind.EXOGENOUS_VARIABLE: "#z",
}


_FROM_PORTABLES = {
    v: k for k, v in _TO_PORTABLES.items()
}


__all__ = (
    "filter_quantities_by_name",
    "TRANSITION_VARIABLE",
    "TRANSITION_SHOCK",
    "ANTICIPATED_SHOCK_VALUE",
    "TRANSITION_STD",
    "MEASUREMENT_VARIABLE",
    "MEASUREMENT_SHOCK",
    "MEASUREMENT_STD",
    "EXOGENOUS_VARIABLE",
    "ANY_VARIABLE",
    "ANY_SHOCK",
    "ANY_SHOCK_OR_SHOCK_VALUE",
    "PARAMETER",
)


@dataclass
class Quantity:
    """
    """
    #[

    id: int | None = None
    human: str | None = None
    kind: QuantityKind = QuantityKind.UNSPECIFIED
    logly: bool | None = None
    description: str | None = None
    entry: int | None = None
    attributes: set[str] | None = None

    def wrap_logly(self, /, ) -> str:
        return wrap_logly(self.human, self.logly, )

    def copy(self, /, ) -> Self:
        """
        Shallow copy of the quantity
        """
        return type(self)(**self.__dict__, )

    def __hash__(self, /, ) -> int:
        return hash(self.__repr__)

    has_attributes = _attributes.has_attributes

    def to_portable(self, /, ) -> _PortableType:
        """
        """
        return (
            self.kind.to_portable(),
            self.human,
            self.logly,
            self.description,
            " ".join(self.attributes, ),
        )

    @classmethod
    def from_portable(klass, portable: _PortableType, /, ) -> Self:
        """
        """
        kind, human, logly, description, attributes = portable
        return klass(
            human=human,
            kind=QuantityKind.from_portable(kind, ),
            logly=logly,
            description=description,
            attributes=set(attributes.split(" ", )),
        )

    #]


def create_name_to_qid(quantities: Iterable[Quantity]) -> dict[str, int]:
    return {
        qty.human: qty.id for qty in quantities
        if qty.id is not None and qty.human is not None
    }


def create_name_to_quantity(quantities: Iterable[Quantity], /, ) -> dict[str, Quantity]:
    return { qty.human: qty for qty in quantities if qty.human is not None }


def create_qid_to_name(quantities: Iterable[Quantity]) -> dict[int, str]:
    return {
        qty.id: qty.human for qty in quantities
        if qty.id is not None and qty.human is not None
    }


def create_qid_to_description(quantities: Iterable[Quantity]) -> dict[int, str]:
    return {
        qty.id: qty.description for qty in quantities
        if qty.id is not None and qty.description is not None
    }


def create_name_to_kind(quantities: Iterable[Quantity]) -> dict[str, str]:
    return {
        qty.human: qty.kind for qty in quantities
        if qty.human is not None and qty.description is not None
    }


def create_name_to_description(quantities: Iterable[Quantity]) -> dict[str, str]:
    return {
        qty.human: qty.description for qty in quantities
        if qty.human is not None and qty.description is not None
    }


def create_qid_to_kind(quantities: Iterable[Quantity]) -> dict[int, QuantityKind]:
    return {
        qty.id: qty.kind for qty in quantities
        if qty.id is not None and qty.kind is not None
    }


def generate_quantities_of_kind(
    quantities: Iterable[Quantity],
    kind: QuantityKind | None = None,
) -> Iterable[Quantity]:
    """
    """
    return ( qty for qty in quantities if kind is None or qty.kind in kind )


def generate_names_of_kind(
    quantities: Iterable[Quantity],
    kind: QuantityKind | None = None,
) -> Iterable[str]:
    """
    """
    return ( qty.human for qty in quantities if kind is None or qty.kind in kind )


def count_quantities_of_kind(
    quantities: Iterable[Quantity],
    kind: QuantityKind | None,
    /,
) -> int:
    """
    """
    return sum(1 for _ in generate_quantities_of_kind(quantities, kind, ))


def generate_qids_by_kind(
    quantities: Iterable[Quantity],
    kind: QuantityKind | None,
) -> Iterable[int | None]:
    """
    """
    return ( qty.id for qty in generate_quantities_of_kind(quantities, kind, ) )


def generate_quantity_names_by_kind(
    quantities: Iterable[Quantity],
    kind: QuantityKind | None,
) -> Iterator[str | None]:
    """
    """
    return ( qty.human for qty in generate_quantities_of_kind(quantities, kind, ) )


def generate_all_quantity_names(quantities: Iterable[Quantity]) -> Iterator[str | None]:
    return ( qty.human for qty in quantities )


def generate_all_qids(quantities: Iterable[Quantity]) -> Iterator[int | None]:
    return ( qty.id for qty in quantities )


def get_max_qid(quantities: Iterable[Quantity]) -> int:
    return max(
        qty.id for qty in quantities
        if qty.id is not None
    )


def create_name_to_logly(quantities: Iterable[Quantity]) -> dict[int, bool]:
    return {
        qty.human: qty.logly for qty in quantities
        if qty.id is not None and qty.logly is not None
    }

def create_qid_to_logly(quantities: Iterable[Quantity]) -> dict[int, bool]:
    return {
        qty.id: qty.logly for qty in quantities
        if qty.id is not None and qty.logly is not None
    }


def generate_logly_indexes(quantities: Iterable[Quantity]) -> Iterator[int]:
    return (
        qty.id for qty in quantities
        if qty.id is not None and qty.logly
    )


def change_logly(
    quantities: Iterable[Quantity],
    new_logly: bool,
    qids: Iterable[int],
    /
) -> Iterable[Quantity]:
    qids = set(qids)
    return [
        qty if qty.id not in qids or qty.logly is None else Quantity(qty.id, qty.human, qty.kind, new_logly)
        for qty in quantities
    ]


def validate_selection_of_quantities(
    allowed_quantities: Iterable[Quantity],
    custom_quantities: Iterable[Quantity] | None,
    /,
) -> tuple[Iterable[Quantity], Iterable[Quantity]]:
    """
    """
    invalid_quantities = list(set(custom_quantities) - set(allowed_quantities)) if custom_quantities is not None else []
    custom_quantities = list(custom_quantities) if custom_quantities is not None else list(allowed_quantities)
    return custom_quantities, invalid_quantities


def lookup_quantities_by_name(
    quantities: Iterable[Quantity],
    custom_names: Iterable[str],
    /,
) -> tuple[tuple[Quantity, ...], tuple[str, ...]]:
    """
    Lookup quantities by name, and return a list of quantities and a list
    of invalid names
    """
    custom_names = list(custom_names)
    name_to_quantity  = create_name_to_quantity(quantities, )
    custom_quantities = tuple(
        name_to_quantity[n]
        for n in custom_names if n in name_to_quantity
    )
    invalid_names = tuple(
        n for n in custom_names if n not in name_to_quantity
    )
    return custom_quantities, invalid_names


def lookup_qids_by_name(
    quantities: Iterable[Quantity],
    custom_names: Iterable[str],
    /,
) -> tuple[tuple[int, ...], tuple[str, ...]]:
    """
    Lookup quantities by name, and return a list of quantities and a list
    of invalid names
    """
    custom_names = tuple(custom_names)
    name_to_qid  = create_name_to_qid(quantities)
    custom_qids = tuple(
        name_to_qid[n]
        for n in custom_names if n in name_to_qid
    )
    invalid_names = tuple(
        n for n in custom_names if n not in name_to_qid
    )
    return custom_qids, invalid_names


def filter_quantities_by_name(
    quantities: Iterable[Quantity],
    /,
    include_names: Iterable[str] | None = None,
    exclude_names: Iterable[str] | None = None,
) -> Iterable[Quantity]:
    """
    """
    include_names = set(include_names) if include_names is not None else None
    exclude_names = set(exclude_names) if exclude_names is not None else None
    inclusion_test = lambda name: include_names is None or name in include_names
    exclusion_test = lambda name: exclude_names is None or name not in exclude_names
    return [ qty for qty in quantities if inclusion_test(qty.human) and exclusion_test(qty.human) ]


def generate_where_logly(
    qids: Iterable[int],
    qid_to_logly: dict[int, bool],
    /,
) -> Iterable[int]:
    """
    """
    return (i for i, qid in enumerate(qids, ) if qid_to_logly.get(qid, False))


def wrap_logly(name, logly: bool = True, /, ) -> str:
    return f"log({name})" if logly else name


def check_unique_names(quantities: Iterable[Quantity], /, ) -> None:
    """
    """
    #[
    name_counter = _co.Counter(q.human for q in quantities)
    if any(c>1 for c in name_counter.values()):
        duplicates = [ n for n, c in name_counter.items() if c>1 ]
        raise _wrongdoings.IrisPieError(
            ["These names are declared multiple times"] + duplicates
        )
    #]


def reorder_by_kind(quantities: Iterable[Quantity], /) -> Iterable[Quantity]:
    return tuple(sorted(quantities, key=lambda x: (x.kind.value, x.entry)))


def stamp_id(quantities: Iterable[Quantity], /) -> None:
    """
    """
    for i, q in enumerate(quantities, ):
        q.id = i


def to_portable(quantities: Iterable[Quantity], /, ) -> tuple[_PortableType, ...]:
    """
    """
    #[
    portable = []
    for kind in _TO_PORTABLES.keys():
        portable += [ i.to_portable() for i in quantities if i.kind == kind ]
    return tuple(portable)
    #]


def from_portable(portables: Iterable[_PortableType], /, ) -> tuple[Quantity, ...]:
    """
    """
    return tuple( Quantity.from_portable(i) for i in portables )


class AccessQuantitiesProtocol(Protocol, ):
    """
    """
    #[

    def _access_quantities(self, /, ) -> Iterable[Quantity, ...]: ...

    #]


class Mixin:
    r"""
    Mixin for objects with quantities
    """
    #[

    def _access_quantities(self, ) -> Iterable[Quantity]:
        r"""
        Default implementation of the _access_quantities method
        """
        return self._invariant.quantities

    def get_quantities(
        self: AccessQuantitiesProtocol,
        kind: QuantityKind | None = None,
    ) -> Iterable[Quantity]:
        """
        """
        quantities = self._access_quantities()
        return tuple(
            generate_quantities_of_kind(quantities, kind, )
            if kind else quantities
        )

    def get_names(
        self,
        kind: QuantityKind | None = None,
    ) -> tuple[str, ...]:
        """
        """
        return tuple(q.human for q in self.get_quantities(kind=kind, ))

    def get_log_status(
        self,
        **kwargs,
    ) -> dict[str, bool]:
        """
        """
        quantities = self._access_quantities()
        return {
            qty.human: qty.logly
            for qty in quantities
            if qty.kind in QuantityKind.LOGGABLE_VARIABLE
        }

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        """
        Create a dictionary mapping quantity names to quantity ids
        """
        quantities = self._access_quantities()
        return create_name_to_qid(quantities, )

    def create_qid_to_name(self, /, ) -> dict[int, str]:
        """
        Create a dictionary mapping quantity ids to quantity names
        """
        quantities = self._access_quantities()
        return create_qid_to_name(quantities, )

    def create_qid_to_kind(self, /, ) -> dict[int, str]:
        """
        Create a dictionary mapping quantity ids to quantity kinds
        """
        quantities = self._access_quantities()
        return create_qid_to_kind(quantities, )

    def create_qid_to_description(self, /, ) -> dict[int, str]:
        """
        Create a dictionary mapping quantity ids to quantity descriptions
        """
        quantities = self._access_quantities()
        return create_qid_to_description(quantities, )

    def create_name_to_description(self, ) -> dict[str, str]:
        """
        """
        quantities = self._access_quantities()
        return create_name_to_description(quantities, )

    def create_qid_to_logly(self, /, ) -> dict[int, bool]:
        """
        Create a dictionary mapping from quantity id to quantity log-status
        """
        quantities = self._access_quantities()
        return create_qid_to_logly(quantities, )

    def get_logly_indexes(self, /) -> tuple[int, ...]:
        """
        Create a tuple of indexes for logly quantities
        """
        quantities = self._access_quantities()
        return tuple(generate_logly_indexes(quantities, ))

    #]






"""
Deep copy mixin
"""


#[
from __future__ import annotations

from typing import (Self, )
import copy as _cp
#]


class Mixin:
    """
    """
    #[
    def copy(self, ) -> Self:
        """
        """
        return _cp.deepcopy(self)
    #]




r"""
"""


#[

from __future__ import annotations

import numpy as _np

from ..dates import Period
from ..dataslates import Dataslate
from ..fords import simulators as _simulators
from ..frames import SingleFrame

from ..progress_bars import ProgressBar

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .red_vars import RedVAR
    from typing import Iterable, Callable, Literal

#]


_rng = _np.random.default_rng()
_standard_normal = _rng.standard_normal


class Inlay:
    r"""
    """
    #[
    def simulate(
        self,
        input_db: Databox,
        span: Iterable[Period],
        #
        residuals_from_data: bool = True,
        deviation: bool = False,
        progress_bar_settings: dict = dict(title="Simulating RedVAR", ),
        **kwargs,
    ) -> Databox:
        r"""
        """
        return _simulate(
            self, input_db, span,
            draw_residuals=None,
            residuals_from_data=residuals_from_data,
            deviation=deviation,
            progress_bar_settings=progress_bar_settings,
            **kwargs,
        )

    def resample(
        self,
        input_db: Databox,
        span: Iterable[Period],
        method: Literal["monte_carlo", "bootstrap", "wild_bootstrap", ],
        progress_bar_settings: dict = dict(title="Resampling RedVAR", ),
        **kwargs,
    ) -> Databox:
        r"""
        """
        return _simulate(
            self, input_db, span,
            draw_residuals=_RESAMPLE_SIMULATOR_DISPATCH[method],
            residuals_from_data=True,
            deviation=False,
            progress_bar_settings=progress_bar_settings,
            **kwargs,
        )

    #]

def _simulate(
    self: RedVAR,
    input_db: Databox,
    span: Iterable[Period],
    #
    residuals_from_data: bool,
    draw_residuals: Callable | None,
    progress_bar_settings: dict,
    #
    target_db: Databox | None = None,
    num_variants: int | None = None,
    remove_initial: bool = False,
    prepend_input: bool = False,
    deviation: bool = False,
    show_progress: bool = False,
) -> Databox:
    r"""
    """
    span = tuple(span)
    num_variants = self.resolve_num_variants_in_context(num_variants, )
    needs_exogenous_impact = not deviation and self.has_exogenous
    needs_draw_residuals = draw_residuals is not None
    ignore_shocks = not residuals_from_data
    slatable = self.slatable_for_simulate(
        residuals_from_data=residuals_from_data,
    )
    dataslate = Dataslate.from_databox_for_slatable(
        slatable, input_db, span,
        num_variants=num_variants,
    )
    frame = SingleFrame(start=span[0], end=span[-1], )
    frame.resolve_columns(dataslate.start, )
    #
    zipped = zip(
        range(num_variants, ),
        self.iter_variants(),
        dataslate.iter_variants(),
    )
    #
    # Simulate each variant
    #=======================================================================
    progress_bar = ProgressBar(
        num_steps=num_variants,
        show_progress=show_progress,
        **progress_bar_settings,
    )
    for vid, model_v, dataslate_v in zipped:
        #
        # Resample residuals
        if needs_draw_residuals:
            draw_residuals(model_v, dataslate_v, )
        #
        # Calculate the impact of exogenous variables
        exogenous_impact = None
        if needs_exogenous_impact:
            exogenous_impact = _simulate_exogenous_impact(model_v, dataslate_v, )
        #
        # Run flat first-order simulation
        _simulators.simulate_flat(
            model_v, dataslate_v, frame,
            deviation=deviation,
            ignore_shocks=ignore_shocks,
            exogenous_impact=exogenous_impact,
        )
        progress_bar.increment()
    #=======================================================================
    #
    if remove_initial:
        dataslate.remove_initial()
    #
    # Convert all variants of the dataslate to a databox
    output_db = dataslate.to_databox()
    if prepend_input:
        output_db.prepend(input_db, span[0]-1, )
    #
    # Add to target databox
    if target_db is not None:
        output_db = target_db | output_db
    #
    return output_db


def _simulate_exogenous_impact(
    model_v: RedVAR,
    dataslate_v: Dataslate,
) -> _np.ndarray:
    r"""
    """
    #[
    B = model_v.get_system_matrices().B
    exogenous_qids = list(model_v.get_exogenous_qids())
    data_array = dataslate_v.get_data_variant()
    x_array = data_array[exogenous_qids, :]
    exogenous_impact = B @ x_array
    return exogenous_impact
    #]


def _resample_residuals_by_wild_bootstrap(
    model_v: RedVAR,
    dataslate_v: Dataslate,
    random_factor_generator: Callable[[tuple[int, int]], _np.ndarray] = _standard_normal,
) -> None:
    r"""
    """
    #[
    order = model_v.order
    residual_qids = model_v.get_residual_qids()
    data_array = dataslate_v.get_data_variant()
    residual_array = data_array[residual_qids, order:]
    num_periods = residual_array.shape[1]
    size = (1, num_periods, )
    factors = random_factor_generator(size, )
    data_array[residual_qids, order:] = factors * residual_array
    #]


def _resample_residuals_by_bootstrap(
    model_v: RedVAR,
    dataslate_v: Dataslate,
) -> None:
    r"""
    """
    #[
    ...
    #]


def _resample_residuals_by_monte_carlo(
    model_v: RedVAR,
    dataslate_v: Dataslate,
) -> None:
    r"""
    """
    #[
    ...
    #]


_RESAMPLE_SIMULATOR_DISPATCH = {
    "monte_carlo": _resample_residuals_by_monte_carlo,
    "bootstrap": _resample_residuals_by_bootstrap,
    "wild_bootstrap": _resample_residuals_by_wild_bootstrap,
}





"""
Resolve substitutions in model equations
"""


#[
from __future__ import annotations

import re as _re
import operator as _op
#]


def resolve_substitutions(
    parsed: dict[str, tuple],
    where_substitute: list[str],
) -> dict[str, tuple]:
    r"""
    """
    source = parsed.get("substitutions", None)
    if not source:
        return parsed
    definitions = _define_substitutions(source, )
    where_substitute = set(where_substitute).intersection(parsed.keys())
    subs_pattern = _re.compile(r"\$(" + "|".join(lhs for lhs in definitions.keys()) + r")\$")
    def replace(match, ) -> str:
        return definitions[match.group(1)]
    def make_substitutions(source: str, ) -> str:
        return _re.sub(subs_pattern, replace, source)
    for wh in where_substitute:
        parsed[wh] = [
            (label, (make_substitutions(dynamic), make_substitutions(steady)), attributes)
            for label, (dynamic, steady), attributes in parsed[wh]
        ]
    return parsed


def _define_substitutions(substitutions: list, ) -> dict[str, str]:
    return dict(_separate_lhs_rhs(s[1][0]) for s in substitutions)


_SUBS_NAME = _re.compile(r"\w+", )


def _separate_lhs_rhs(subs_string: str, ) -> tuple[str, str]:
    subs_string = subs_string.replace(":=", "=", )
    lhs, rhs = subs_string.split("=", maxsplit=1, )
    return lhs, rhs





"""
"""


#[
from __future__ import annotations

from typing import (Any, Self, Iterable, Iterator, TypeVar, Protocol, )
from numbers import (Number, )
import numpy as _np
import functools as _ft

from .conveniences import iterators as _iterators
#]


_T = TypeVar("T")


class HasVariantsProtocol:
    """
    """
    #[

    _invariant: Any | None
    _variants: list | None

    #]


class Mixin:
    """
    Mixin for handling multiple alternative variants of a single model object
    """
    #[

    @classmethod
    def skeleton(
        klass,
        other,
    ) -> Self:
        r"""
        """
        return klass(invariant=other._invariant, )

    def select_variants(
        self,
        selectors: Iterable[bool] | None = None,
        indices: Iterable[int] | None = None,
    ) -> None:
        if (selectors is None) + (indices is None) != 1:
            raise Exception("Either selectors or indices must be specified, but not both.")
        if selectors is not None:
            new_variants = [ v for v, s in zip(self._variants, selectors) if s ]
        if indices is not None:
            new_variants = [ self._variants[i] for i in indices ]
        self._variants = new_variants

    @property
    def num_variants(self, ) -> int:
        """
        Number of alternative variants within this model
        """
        return len(self._variants)

    @property
    def is_singleton(self, ) -> bool:
        """
        True for models with only one variant
        """
        return is_singleton(self.num_variants, )

    def new_with_shallow_variants(self, ) -> Self:
        """
        Create a new model with pointers to invariant and variants of this model
        """
        new = type(self)()
        new._invariant = self._invariant
        new._variants = [ v for v in self._variants ]
        return new

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        Alter (expand, shrink) the number of alternative parameter variants in this model object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )

    def shrink_num_variants(self, new_num: int, ) -> None:
        r"""
        """
        if new_num < 1:
            raise Exception('Number of variants must be one or more')
        self._variants = (
            self._variants[0:new_num]
            if new_num < self.num_variants
            else self._variants
        )

    def expand_num_variants(self, new_num: int, ) -> None:
        r"""
        """
        if new_num < 1:
            raise Exception('Number of variants must be one or more')
        for i in range(self.num_variants, new_num):
            self._variants.append(self._variants[-1].copy(), )

    def resolve_num_variants_in_context(self, custom_num_variants: int | None, ) -> int:
        """
        """
        if custom_num_variants is None:
            return self.num_variants
        else:
            return custom_num_variants

    def broadcast_variants(self, other: Self, ) -> None:
        """
        """
        if self.num_variants < other.num_variants:
            self.expand_num_variants(other.num_variants, )

    def get_variant(
        self,
        vids: Iterable[int] | int | slice | EllipsisType,
    ) -> Self:
        new = self.skeleton(self, )
        new._variants = [ self._variants[i] for i in _resolve_vids(self, vids, ) ]
        return new

    def iter_variants(self, ) -> Iterator[Self]:
        r"""
        """
        return _iterators.exhaust_then_last(self, )

    def iter_extract_variants(self, ) -> Iterator:
        """
        """
        return _iterators.exhaust_then_last(self._variants, )

    def iter_own_variants(self, ) -> Iterator[Self]:
        r"""
        Iterate over alternative variants of this object
        """
        for v in self._variants:
            new = self.skeleton(self, )
            new._variants = [v, ]
            yield new

    def __iter__(self, ) -> Iterator[Self]:
        r"""
        Iterate over alternative variants of this object
        """
        return self.iter_own_variants()

    def repack_singleton(
        self: HasVariantsProtocol,
        anything: list[_T] | _T,
        /,
    ) -> list[_T]:
        """
        """
        return [anything, ] if not isinstance(anything, list) else anything

    def unpack_singleton(
        self: HasVariantsProtocol,
        anything: list[_T] | _T,
        **kwargs,
    ) -> list[_T] | _T:
        """
        """
        return unpack_singleton(
            anything,
            self.is_singleton,
            **kwargs,
        )

    def unpack_singleton_in_dict(
        self: HasVariantsProtocol,
        anything_in_dict: dict[str, list[_T] | _T],
        **kwargs,
    ) -> dict[str, _T | list[_T]]:
        """
        """
        return unpack_singleton_in_dict(
            anything_in_dict,
            self.is_singleton,
            **kwargs,
        )
    #]


def _resolve_vids(
    self,
    vids: Iterable[int] | int | slice | EllipsisType,
    /,
) -> Iterable[int]:
    """
    """
    #[
    if isinstance(vids, Number):
        return (int(vids), )
    elif vids is ...:
        return range(self.num_variants, )
    elif isinstance(vids, slice):
        return range(*vids.indices(self.num_variants, ), )
    else:
        return tuple(int(i) for i in vids)
    #]


def iter_variants(anything: _T, ) -> Iterator[_T]:
    """
    """
    if hasattr(anything, "iter_variants", ):
        return anything.iter_variants()
    elif isinstance(anything, _np.ndarray, ) and anything.ndim:
        return _iterators.exhaust_then_last(anything.reshape(anything.shape[0], -1, ).T)
    elif isinstance(anything, list, ):
        return _iterators.exhaust_then_last(anything, )
    else:
        return _iterators.exhaust_then_last([anything, ], )


def iter_own_variants(anything: _T, ) -> Iterator[_T]:
    """
    """
    if hasattr(anything, "iter_own_variants", ):
        return anything.iter_own_variants()
    elif isinstance(anything, _np.ndarray, ) and anything.ndim:
        return anything.reshape(anything.shape[0], -1, ).T
    elif isinstance(anything, list, ):
        return anything
    else:
        return [anything, ]


def unpack_singleton(
    anything: list[_T],
    is_singleton: bool | None = True,
    unpack_singleton: bool = True,
) -> _T | list[_T]:
    """
    """
    if not unpack_singleton:
        return anything
    if is_singleton is None:
        is_singleton = len(anything) == 1
    return anything[0] if is_singleton else anything


shadow_unpack_singleton = unpack_singleton


def unpack_singleton_in_dict(
    anything_in_dict: dict[str, list[_T] | _T],
    is_singleton: bool,
    /,
    unpack_singleton: bool = True,
) -> dict[str, _T | list[_T]]:
    """
    """
    if is_singleton and unpack_singleton:
        return {
            k: shadow_unpack_singleton(v, )
            for k, v in anything_in_dict.items()
        }
    else:
        return anything_in_dict


def is_singleton(num_variants: int) -> bool:
    """
    """
    return num_variants == 1


def unpack_singleton_decorator(func: Callable, ):
    """
    """
    #[
    @_ft.wraps(func, )
    def _wrapper(self, *args, **kwargs, ):
        unpack_singleton = kwargs.pop("unpack_singleton", True)
        output = func(self, *args, **kwargs)
        return self.unpack_singleton(output, unpack_singleton=unpack_singleton, )
    return _wrapper
    #]




r"""
"""


#[

from __future__ import annotations

import time as _tm
import sys
from typing import Any

#]


__all__ = (
    "ProgressBar",
)


_DEFAULT_EMPTY_CHAR = "◻︎"
_DEFAULT_FILL_CHAR = "◼︎"
_DEFAULT_BAR_LENGTH = 50


class ProgressBar:
    r"""
    """

    def __init__(
        self,
        num_steps: int,
        #
        title: str = "",
        bar_length: int = _DEFAULT_BAR_LENGTH,
        fill_char: str = _DEFAULT_FILL_CHAR,
        empty_char: str = _DEFAULT_EMPTY_CHAR,
        show_progress: bool = True,
        progress_format: str = "{title} {bar} {percent:.2f}%",
        final_format: str = "{title} {bar} {time_elapsed:.4f}s",
        output_stream: Any = sys.stdout,
        start: bool = True,
        finish_when_complete: bool = True,
    ) -> None:
        r"""
        """
        self.num_steps = num_steps
        self.title = title
        self.bar_length = bar_length
        self.current_step = 0 
        self.fill_char = fill_char[0]
        self.empty_char = empty_char[0]
        self.show_progress = show_progress
        self.progress_format = progress_format
        self.final_format = final_format
        self.output_stream = output_stream
        self.time_started = None
        self.time_finished = None
        self.finish_when_complete = finish_when_complete
        if start:
            self.start()

    @property
    def is_complete(self) -> bool:
        r"""
        """
        return self.current_step >= self.num_steps

    def start(self, ) -> None:
        r"""
        """
        self.time_started = _tm.time()
        self.update(0, )

    def update(self, step: int, ) -> None:
        r"""
        """
        self.current_step = step
        if self.show_progress:
            self._print_bar()
        if self.finish_when_complete and self.is_complete:
            self.finish()

    def increment(self, step: int = 1, ) -> None:
        r"""
        """
        self.update(self.current_step + step, )

    def finish(self, ) -> None:
        r"""
        """
        self.time_finished = _tm.time()
        if self.show_progress:
            self._print_final()

    def _print_bar(self, ) -> None:
        r"""
        """
        context = self._eval_context(_tm.time(), )
        progress_line = self.progress_format.format(**context, )
        print(progress_line, end="\r", file=self.output_stream, flush=True, )

    def _print_final(self, ) -> None:
        r"""
        """
        context = self._eval_context(self.time_finished, )
        final_line = self.final_format.format(**context, )
        print(final_line, file=self.output_stream, flush=True, )

    def _eval_context(self, time, ) -> dict[str, Any]:
        r"""
        """
        filled_length = int(self.bar_length * self.current_step // self.num_steps)
        return dict(
            title=self.title,
            percent=(self.current_step / self.num_steps) * 100,
            bar=self.fill_char * filled_length + self.empty_char * (self.bar_length - filled_length),
            time_elapsed=time - self.time_started,
        )





#[

import importlib.metadata as _md
import re as _re
import functools as _ft
import warnings as _wa

_BANNER = {
    "de": (
        "\n\n"
        "=================================================================\n"
        " You are using IrisPie Developer Edition.\n"
        " This edition is intended for development and testing only.\n"
        " It may be used solely by registered developers.\n"
        " Contact info@ogresearch.com for more information.\n"
        "=================================================================\n\n"
    ),
    "ce": (
        "\n\n"
        "=================================================================\n"
        " You are using IrisPie Community Edition.\n"
        " Free for personal, educational, and non-commercial use only.\n"
        " Registration is required for commercial or institutional use.\n"
        " Contact info@ogresearch.com for more information.\n"
        "=================================================================\n\n"
    ),
    "re": (
        "\n\n"
        "=================================================================\n"
        " You are using IrisPie Registered Edition.\n"
        " Licensed for use by organizations that have completed\n"
        " registration with OGResearch.\n"
        " Internal use and redistribution within the organization\n"
        " are permitted. External redistribution is prohibited.\n"
        " Contact info@ogresearch.com for more information.\n"
        "=================================================================\n\n"
    ),
    "pe": (
        "\n\n"
        "=================================================================\n"
        " You are using IrisPie Private Edition.\n"
        " This edition may only be used internally by staff members\n"
        " of OGResearch or by approved contractors and affiliated\n"
        " institutions.\n"
        "=================================================================\n\n"
    ),
}

distribution_generator = (
    i for i in _md.distributions()
    if _re.match("irispie-[dcrp]e$", i.name)
)
distribution = next(distribution_generator, None, )

if not distribution:
    raise Exception("Cannot determine the irispie distribution", )

edition = distribution.name[-2:]
version = distribution.version + "-" + edition
__version__ = version
__doc__ = distribution.metadata["description"]

irispie_edition_warning = _ft.partial(_wa.warn, _BANNER[edition], UserWarning, )
irispie_edition_warning()

#]


#[

def print_readme():
    print(__doc__)

def min_version_required(
    min_version_string: str,
):
    """
    """
    current_version = _convert_version(__version__, )
    minimum_version = _convert_version(min_version_string, )
    if current_version < minimum_version:
        raise Exception(
            f"Current version of irispie ({__version__}) is less than the minimum version required ({min_version_string})"
        )

min_irispie_version_required = min_version_required

def _convert_version(version_str: str) -> tuple[int, ...]:
    return tuple(int(s) for s in version_str.split("."))

#]


from .dates import *
from .dates import __all__ as dates_all

from .series import *
from .series import __all__ as series_all

from .ez_plotly import *
from .ez_plotly import __all__ as ez_plotly_all

from .databoxes import *
from .databoxes import __all__ as databoxes_all

from .dataslates import *
from .dataslates import __all__ as dataslates_all

from .chartpacks.main import *
from .chartpacks.main import __all__ as chartpacks_all

from .sources import *
from .sources import __all__ as sources_all

from .simultaneous import *
from .simultaneous import __all__ as simultaneous_all

from .stackers import *
from .stackers import __all__ as stackers_all

from .fords import *
from .fords import __all__ as fords_all

from .red_vars import *
from .red_vars import __all__ as red_vars_all

from .quantities import *
from .quantities import __all__ as quantities_all

from .equations import *
from .equations import __all__ as equations_all

from .sequentials import *
from .sequentials import __all__ as sequentials_all

from .explanatories import *
from .explanatories import __all__ as explanatories_all

from .plans import *
from .plans import __all__ as plans_all

from .namings import *
from .namings import __all__ as namings_all

from .file_io import *
from .file_io import __all__ as file_io_all

from .portables import *
from .portables import __all__ as portables_all

from .progress_bars import *
from .progress_bars import __all__ as progress_bars_all


__all__ = (
    *dates_all,
    *series_all,
    *ez_plotly_all,
    *databoxes_all,
    *dataslates_all,
    *sources_all,
    *simultaneous_all,
    *stackers_all,
    *fords_all,
    *red_vars_all,
    *quantities_all,
    *equations_all,
    *sequentials_all,
    *explanatories_all,
    *plans_all,
    *namings_all,
    *file_io_all,
    *portables_all,
    *progress_bars_all,
    "min_version_required",
    "min_irispie_version_required",
    "__version__",
)




"""
Simultaneous models
"""

# Attributes to be exposed at the package level

from .main import *
from .main import __all__ as main_all

__all__ = []
__all__.extend(main_all, )


# Attributes to be exposed within the package

from ._invariants import (
    anticipated_shock_name_from_transition_shock_name,
    is_anticipated_shock_name,
)




"""
Documark categories for the Series class
"""


CATEGORIES = {
    "constructor": "Constructing new time series",
    "conversion": "Converting time series frequency",
    "information": "Getting information about time series",
    "manipulation": "Manipulating time series values",
    "comparison": "Comparing time series",
    "multiple": "Combining multiple time series",
    "homogenizing": "Homogenizing and extrapolating time series",
    "filtering": "Filtering time series",
    "moving": "Applying moving window functions",
    "temporal_change": "Calculating temporal change",
    "temporal_change_conversion": "Converting measures of temporal change",
    "temporal_cumulation": "Calculating temporal cumulation",
    "indexing": "Indexing time series",
    "serialization": "Serializing and deserializing time series",
}




"""
"""


#[

from __future__ import annotations

from typing import Literal, Callable
from collections.abc import Iterable
from numbers import Real
import functools as _ft
import json as _js
import numpy as _np
import documark as _dm

from .. import equations as _equations
from .. import quantities as _quantities
from ..quantities import QuantityKind
from .. import has_variants as _has_variants
from ..has_variants import unpack_singleton_decorator as _unpack_singleton
from ..series import main as _series
from ..incidences import main as _incidence
from ..databoxes.main import Databox
from ..fords.solutions import EigenvalueKind, Solution
from ..fords import descriptors as _descriptors


from . import _flags

#]


def _cast_as_output_type(func: Callable, ):
    r"""
    Decorator to cast the output of a function to a specified type.
    """
    #[
    @_ft.wraps(func)
    def _wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        output_type = kwargs.get("output_type", Databox, )
        return output_type(output, ) if output_type else output
    return _wrapper
    #]


def _unpack_singleton_in_dict(func: Callable, ):
    r"""
    Decorator to unpack output from a singleton model
    """
    #[
    @_ft.wraps(func)
    def _wrapper(self, *args, **kwargs, ):
        unpack_singleton = kwargs.pop("unpack_singleton", True)
        output = func(self, *args, **kwargs)
        return self.unpack_singleton_in_dict(output, unpack_singleton=unpack_singleton, )
    return _wrapper
    #]


class Inlay(
    _quantities.Mixin,
):
    """
    Frontend getter methods for Simultaneous objects
    """
    #[

    @_dm.reference(category="information", )
    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_steady_levels(
        self,
        unpack_singleton: bool = True,
        kind: _quantities.QuantityKind | None = None,
        **kwargs,
    ) -> dict[str, list[Real] | Real]:
        """
················································································

==Get steady-state levels of variables==

    steady_levels = self.get_steady_levels(
        *,
        round: int = None,
        unpack_singleton: bool = True,
        output_typ: type = Databox,
    )


### Input arguments ###


???+ input "self"
    `Simultaneous` model object with a valid first-order solution.

???+ input "round"
    Number of decimal places to round the output values.

???+ input "unpack_singleton"
    If `True`, unpack singleton lists and return the single element. If
    `False`, return the list as is.

???+ input "output_type"
    Cast output as this type; the constructor for this type must accept a
    dictionary.


### Returns ###


???+ returns "steady_levels"
    Databox with the steady-state levels for each variable.

......................................................................
        """
        kind = kind if kind is not None else _resolve_steady_kind(**kwargs, )
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, kind, )
        return self._get_values_as_dict("levels", qids, **kwargs, )

    @_dm.reference(category="information", )
    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_steady_changes(
        self,
        unpack_singleton: bool = True,
        kind: _quantities.QuantityKind | None = None,
        **kwargs,
    ) -> dict[str, list[Real] | Real]:
        """
················································································

==Get steady-state changes of variables==

    steady_changes = self.get_steady_changes(
        *,
        round: int = None,
        unpack_singleton: bool = True,
        output_typ: type = Databox,
    )


### Input arguments ###


???+ input "self"
    `Simultaneous` model object with a valid first-order solution.

???+ input "round"
    Number of decimal places to round the output values.

???+ input "unpack_singleton"
    If `True`, unpack singleton lists and return the single element. If
    `False`, return the list as is.

???+ input "output_type"
    Cast output as this type; the constructor for this type must accept a
    dictionary.


### Returns ###


???+ returns "steady_changes"
    Databox with the steady-state changes for each variable, i.e. first
    differences or gross rates of change depending on the log status of
    each variable.

......................................................................
        """
        kind = kind if kind is not None else _resolve_steady_kind(**kwargs, )
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, kind, )
        return self._get_values_as_dict("changes", qids, **kwargs, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_steady(
        self,
        unpack_singleton: bool = True,
        kind: _quantities.QuantityKind | None = None,
        **kwargs,
    ) -> dict[str, list[tuple[Real, Real]] | tuple[Real, Real]]:
        """
        """
        kind = kind if kind is not None else _resolve_steady_kind(**kwargs, )
        qids = tuple(_quantities.generate_qids_by_kind(self._invariant.quantities, kind, ))
        levels = self._get_values_as_dict("levels", qids, **kwargs, )
        changes = self._get_values_as_dict("changes", qids, **kwargs, )
        return {
            k: list(zip(levels[k], changes[k]))
            for k in levels.keys()
        }

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_parameters(
        self,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Real]:
        """
        """
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.PARAMETER, )
        return self._get_values_as_dict("levels", qids, **kwargs, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_stds(
        self,
        kind: _quantities.QuantityKind | None = None,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Real]:
        """
        """
        std_qids = self._get_std_qids(kind=kind, )
        return self._get_values_as_dict("levels", std_qids, **kwargs, )

    def _get_std_qids(
        self,
        kind: _quantities.QuantityKind | None = None,
    ) -> tuple[int, ...]:
        """
        """
        default_kind = _quantities.QuantityKind.ANY_STD
        kind = default_kind & kind if kind is not None else default_kind
        return _quantities.generate_qids_by_kind(self._invariant.quantities, kind, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_parameters_stds(
        self,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Real]:
        """
        """
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.PARAMETER_OR_STD, )
        return self._get_values_as_dict("levels", qids, **kwargs, )

    def get_initials(
        self,
    ) -> _incidence.Tokens:
        """
        Get required list of initial conditions
        """
        # Get the tokens of state vector with actual incidence in the
        # model, and lag them by one period to get initial conditions
        initial_tokens = [
            _incidence.Token(t.qid, t.shift-1)
            for t in self._invariant.dynamic_descriptor.solution_vectors.get_initials()
        ]
        qid_to_name = self.create_qid_to_name()
        qid_to_logly = {} # [^1]
        return _incidence.print_tokens(initial_tokens, qid_to_name, qid_to_logly, )
        # [^1] Do not wrap initial conditions in log(...)

    def _get_dynamic_solution_vectors(self, ) -> _descriptors.SolutionVectors:
        r"""
        FordSimulatableProtocol
        """
        return self._invariant.dynamic_descriptor.solution_vectors

    def get_solution_vectors(self, ) -> _descriptors.HumanSolutionVectors:
        """
        Get the solution vectors of the model
        """
        qid_to_name = self.create_qid_to_name()
        qid_to_logly = self.create_qid_to_logly()
        return _descriptors.HumanSolutionVectors(
            self._dynamic_solution_vectors,
            qid_to_name,
            qid_to_logly,
        )

    def get_solution(
        self,
        #
        unpack_singleton: bool = True,
    ) -> Solution | list[Solution]:
        """
        """
        solution_matrices = [ i for i in self.iter_solution() ]
        return self.unpack_singleton(solution_matrices, unpack_singleton=unpack_singleton, )

    def iter_solution(self, ) -> Iterable[Solution]:
        """
        """
        for v in self._variants:
            yield v.solution

    def _gets_solution(
        self,
        deviation: bool = False,
    ) -> Solution:
        r"""
        """
        solution = self._variants[0].solution
        if deviation:
            solution = solution.create_deviation_solution()
        return solution

    def iter_std_name_to_value(self, ) -> Iterable[dict[str, Real]]:
        """
        """
        qid_to_name = self.create_qid_to_name()
        std_qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.ANY_STD, )
        for v in self._variants:
            std_qid_to_value = v.retrieve_levels_as_dict(std_qids, )
            yield {
                qid_to_name[qid]: value
                for qid, value in std_qid_to_value.items()
            }

    @_dm.reference(category="information", )
    def get_equations(
        self,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[str]:
        r"""
...................................................................................

==Get the model equations==

    equations = self.get_equations(
        kind=None,
    )


### Input arguments ###

???+ input "self"
    A `Simultaneous` model object.

???+ input "kind"
    The kind of equations to retrieve. If `None`, all equations are returned.


### Returns ###

???+ returns "equations"
    A tuple of strings representing the equations of the model, with dynamic
    and steady equations concatenated using "!!" if they differ.

...................................................................................
        """
        def concatenate(dynamic: str, steady: str) -> str:
            return f"{dynamic} !! {steady}" if steady != dynamic else dynamic
        zipper = zip(
            self.get_dynamic_equations(kind=kind, ),
            self.get_steady_equations(kind=kind, ),
        )
        return tuple(
            concatenate(dynamic, steady)
            for dynamic, steady in zipper
        )

    get_human_equations = get_equations

    def get_dynamic_equations(
        self,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        r"""
        """
        return tuple(
            i.human
            for i in self.get_dynamic_equation_objects(kind=kind, )
        )

    def get_steady_equations(
        self,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        r"""
        """
        return tuple(
            i.human
            for i in self.get_steady_equation_objects(kind=kind, )
        )

    def get_dynamic_equation_objects(
        self,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        return tuple(
            _equations.generate_equations_of_kind(self._invariant.dynamic_equations, kind)
            if kind else self._invariant.dynamic_equations
        )

    def get_steady_equation_objects(
        self,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        return tuple(
            _equations.generate_equations_of_kind(self._invariant.steady_equations, kind)
            if kind else self._invariant.steady_equations
        )

    def get_flags(
        self,
    ) -> _flags.Flags:
        return self._invariant._flags

    def get_context(
        self,
    ) -> dict[str, Callable]:
        return self._invariant._context

    def _get_values_as_dict(
        self,
        variant_attr: Literal["levels", "changes"],
        qids: Iterable[int],
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        custom_round = kwargs.pop("round", None, )
        def apply(value: Real) -> Real:
            return (
                round(value, custom_round)
                if custom_round is not None and value is not None
                else value
            )
        #
        qid_to_name = self.create_qid_to_name()
        out_dict = {}
        for qid in qids:
            out_dict[qid_to_name[qid]] = [
                apply(getattr(v, variant_attr, )[qid], )
                for v in self._variants
            ]
        return out_dict

    def get_equation_descriptions(self, ) -> dict[str, str]:
        r"""
        """
        return _equations.create_human_to_description(self._invariant.dynamic_equations, )

    @_unpack_singleton
    def get_eigenvalues(
        self,
        transform: Callable[[Real], Real] = lambda x: x,
        kind: EigenvalueKind | None = EigenvalueKind.ALL,
        unpack_singleton: bool = True,
    ) -> tuple[Real, ...] | list[tuple[Real, ...]]:
        r"""
        Eigenvalues
        """
        if kind is None:
            kind = EigenvalueKind.ALL
        return [
            tuple(transform(e) for e, s in zip(v.solution.eigenvalues, v.solution.eigenvalues_stability, ) if s in kind)
            for v in self._variants
        ]

    @_unpack_singleton
    def get_eigenvalues_stability(
        self,
        kind: EigenvalueKind = EigenvalueKind.ALL,
        unpack_singleton: bool = True,
    ):
        return [ 
            tuple(i for i in v.solution.eigenvalues_stability if i in kind)
            for v in self._variants
        ]

    def _get_variable_stability_for_variant(self, variant, ) -> dict[str, bool]:
        """
        """
        vec = self._invariant.dynamic_descriptor.solution_vectors
        qid_to_name = self.create_qid_to_name()
        return {
            qid_to_name[token.qid]: \
                variant.solution.transition_vector_stability[index] == EigenvalueKind.STABLE
            for index, token in enumerate(vec.transition_variables)
            if token.shift == 0
        } | {
            qid_to_name[token.qid]: \
                variant.solution.measurement_vector_stability[index] == EigenvalueKind.STABLE
            for index, token in enumerate(vec.measurement_variables)
            if token.shift == 0
        }

    @_unpack_singleton
    def get_variable_stability(
        self,
        unpack_singleton: bool = True,
    ) -> dict[str, bool] | list[dict[str, bool]]:
        """
        """
        return [
            self._get_variable_stability_for_variant(v, )
            for v in self._variants
        ]

    def generate_minus_control_quantities(self, ) -> tuple[int]:
        """
        """

    def map_name_to_minus_control_func(self, ) -> tuple[str]:
        """
        """
        minus_control_func = {
            True: lambda x, y: x / y,
            False: lambda x, y: x - y,
            None: lambda x, y: x - y,
        }
        kind = _quantities.ANY_VARIABLE | _quantities.ANY_SHOCK_OR_SHOCK_VALUE
        return {
            q.human: minus_control_func[q.logly]
            for q in self._invariant.quantities
            if q.kind in kind
        }

    #]


# Decorate methods from quantities.Mixin
Inlay.get_log_status = _cast_as_output_type(Inlay.get_log_status, )


def _resolve_steady_kind(
    include_shocks: bool = False,
    **kwargs,
) -> _quantities.QuantityKind:
    """
    """
    #[
    steady_kind = QuantityKind.LOGGABLE_VARIABLE
    if include_shocks:
        steady_kind = steady_kind | QuantityKind.ANY_SHOCK_OR_SHOCK_VALUE
    return steady_kind
    #]




r"""
Main time series class definition
"""


#[

from __future__ import annotations

from numbers import Real
from collections.abc import Iterable, Callable
from typing import Self, Any, TypeAlias, NoReturn, Literal
from types import EllipsisType
import numpy as _np
import operator as _op
import documark as _dm

from ..conveniences import descriptions as _descriptions
from ..conveniences import copies as _copies
from ..conveniences import iterators as _iterators
from ..dates import Period, Span, Frequency, EmptyRanger
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants

from ._functionalize import FUNC_STRING

from . import _indexing
from . import _plotly
from . import _views
from . import _temporal
from . import _filling
from . import _conversions
from . import _hp
from . import _extrapolate
from . import _x13
from . import _moving
from . import _elementwise
from . import _statistics
from . import _jsonables
from . import _lays
from . import _broadcasts

from ._categories import CATEGORIES

#]


#
# The following functions can return either a new time series object if applied
# across variants (axis=1), or a scalar or a list of values if applied across
# time (axis=0)
#


__all__ = (
    "Series",
)


Dates = Period | Iterable[Period] | Span | EllipsisType | None
VariantsRequestType = int | Iterable[int] | slice | None
ShiftType = int | Literal["yoy", "soy", "eopy", "tty", ]
AxisType = Literal[0, 1]


def _get_date_positions(dates, base, num_periods, ):
    pos = tuple(_dates.period_indexes(dates, base, ))
    min_pos = min((x for x in pos if x is not None), default=0)
    max_pos = max((x for x in pos if x is not None), default=0)
    add_before = max(-min_pos, 0)
    add_after = max(max_pos - num_periods + 1, 0)
    pos_adjusted = [
        p + add_before if p is not None else None
        for p in pos
    ]
    return pos_adjusted, add_before, add_after


@_lays.mixin
@_filling.mixin
@_broadcasts.mixin
@_jsonables.mixin
@_dm.reference(
    path=("data_management", "time_series.md", ),
    categories=CATEGORIES,
)
class Series(
    _indexing.Inlay,
    _conversions.Inlay,
    _temporal.Inlay,
    _hp.Inlay,
    _extrapolate.Inlay,
    _moving.Inlay,
    _x13.Inlay,
    _plotly.Inlay,
    _views.Inlay,
    _elementwise.Inlay,
    _statistics.Inlay,
    #
    _descriptions.DescriptionMixin,
    _copies.Mixin,
):
    r"""
················································································

Time series
============

Time `Series` objects represent numerical time series, organized as rows of
observations stored in [`numpy`](https://numpy.org) arrays and time stamped
using [time `Periods`](periods.md). A `Series` object can hold multiple
variants of the data, stored as mutliple columns.

················································································
    """
    #[

    __slots__ = (
        "start",
        "data",
        "data_type",
        "__description__",
    )

    _numeric_format: str = "15g"
    _short_str_format: str = ">15"
    _date_str_format: str = ">12"
    _missing = _np.nan
    _missing_str: str = "⋅"
    _test_missing_period = staticmethod(lambda x: _np.all(_np.isnan(x)))

    @_dm.reference(
        category="constructor",
        call_name="Series",
    )
    def __init__(
        self,
        *,
        num_variants: int = 1,
        data_type: type = _np.float64,
        description: str = "",
        start: Period | None = None,
        start_date: Period | None = None,
        periods: Iterable[Period] | None = None,
        dates: Iterable[Period] | None = None,
        frequency: Frequency | None = None,
        values: Any | None = None,
        func: Callable | None = None,
        populate: bool = True,
    ) -> None:
        """
················································································

==Create a new `Series` object==

    self = Series(
        start=start,
        values=values,
    )


    self = Series(
        periods=periods,
        values=values,
    )


    self = Series(
        periods=periods,
        func=func,
    )


### Input arguments ###

???+ input "start"
    The time [`Period`](periods.md) of the first value in the `values`.

???+ input "periods"
    An iterable of time [`Periods`](periods.md) that will be used to time stamp
    the `values`. The iterable can be e.g. a tuple, a list, a time
    [`Span`](spans.md), or a single time [`Period`](periods.md).

???+ input "values"
    Time series values, supplied either as a single values, a tuple of values,
    or a NumPy array.

???+ input "func"
    A function that will be used to populate the time series; the function
    should not take any input arguments, and should return a single (scalar)
    numerical value; the function will called once for each period and each
    variant.


### Returns ###

???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        self.start = None
        self.data = _np.empty((0, num_variants), dtype=data_type, )
        self.data_type = self.data.dtype
        self.set_description(description, )
        #
        # Legacy support for start_date and dates
        if start is None and start_date is not None:
            start = start_date
        if periods is None and dates is not None:
            periods = dates
        #
        if populate:
            test = tuple(x is not None for x in (start, periods, values, func))
            populator = _SERIES_POPULATOR.get(test, _invalid_constructor)
            populator(
                self,
                start=start,
                periods=periods,
                frequency=frequency,
                values=values,
                func=func,
            )

    def copy(self, ) -> Self:
        r"""
        """
        new = type(self)(populate=False, )
        new.start = None
        if self.start is not None:
            new.start = self.start.copy()
        new.data = _np.array(self.data, )
        new.data_type = self.data_type
        new.set_description(self.get_description(), )
        return new

    @classmethod
    def as_empty(
        klass,
        num_variants: int = 1,
        data_type: type = _np.float64,
        description: str = "",
    ) -> Self:
        r"""
        """
        return klass(
            num_variants=num_variants,
            data_type=data_type,
            description=description,
            populate=False,
        )

    @classmethod
    @_dm.reference(
        category="constructor",
        call_name="Series.from_start_and_array",
    )
    def from_start_and_array(
        klass,
        start: Period,
        array: _np.ndarray,
        description: str = "",
        trim: bool = True,
    ) -> Self:
        r"""
················································································

==Create a new `Series` object from a start period and a numpy array==

    self = Series.from_start_and_array(
        start,
        array,
        description="",
        trim=True,
    )


### Input arguments ###

???+ input "start"
    The time [`Period`](periods.md) of the first value in the `array`.

???+ input "array"
    A numpy array containing the time series values. The array is reshaped to a
    two-dimensional array if needed. The individual rows are expected to
    correspond to a continuous sequence of time periods starting from `start`.

???+ input "description"
    A string description of the time series.

???+ input "trim"
    If `True`, the time series date will be trimmed to remove any leading
    or trailing periods that contain only `nan` values.


### Returns ###

???+ returns "self"
    A new `Series` object initialized with the provided start period and
    numpy array.

................................................................................
        """
        self = klass()
        self.start = start
        self.data = _reshape_numpy_array(array, )
        self.set_description(description, )
        if trim:
            self.trim()
        return self

    def reset(self, ) -> None:
        self.__init__(
            num_variants=self.num_variants,
            data_type=self.data_type,
        )

    def _create_periods_of_missing_values(self, num_rows=0) -> _np.ndarray:
        """
        """
        return _np.full(
            (num_rows, self.shape[1], ),
            _np.nan,
            dtype=self.data_type
        )

    @property
    @_dm.reference(category="property", )
    def shape(self, ) -> tuple[int, int]:
        """==Shape of time series data=="""
        return self.data.shape

    @property
    @_dm.reference(category="property", )
    def num_periods(self, ) -> int:
        """==Number of periods from the first to the last observation=="""
        return self.data.shape[0]

    @property
    @_dm.reference(category="property", )
    def num_variants(self, ) -> int:
        """==Number of variants (columns) within the `Series` object=="""
        return self.data.shape[1]

    @property
    def is_singleton(self, ) -> bool:
        """
        True for time series with only one variant
        """
        return _has_variants.is_singleton(self.num_variants, )

    @property
    @_dm.reference(category="property", )
    def span(self, ):
        """==Time span of the time series=="""
        return Span(self.start, self.end, ) if self.start else ()

    range = span

    @property
    @_dm.reference(category="property", )
    def from_until(self, ):
        """==Two-tuple with the start date and end date of the time series=="""
        return self.start, self.end

    @property
    @_dm.reference(category="property", )
    def periods(self, ) -> tuple[Period, ...]:
        """==N-tuple with the periods from the start period to the end period of the time series=="""
        return tuple(self.span, )

    dates = periods

    @property
    @_dm.reference(category="property", call_name="start", )
    def _start(self):
        """==Start date of the time series=="""
        raise NotImplementedError

    @property
    def start_period(self, ):
        return self.start

    start_date = start_period

    @property
    @_dm.reference(category="property", )
    def end(self):
        """==End period of the time series=="""
        return (
            self.start + self.data.shape[0] - 1
            if self.start else None
        )

    end_period = end
    end_date = end

    @property
    @_dm.reference(category="property", )
    def frequency(self, ):
        """==Date frequency of the time series=="""
        return (
            self.start.frequency
            if self.start is not None
            else Frequency.UNKNOWN
        )

    @property
    @_dm.reference(category="property", )
    def is_empty(self, ) -> bool:
        """==True if the time series is empty=="""
        return not self.data.size

    @property
    @_dm.reference(category="property", )
    def has_missing(self, ):
        """==True if the time series is non-empty and contains in-sample missing values=="""
        return bool((not self.is_empty) and _np.isnan(self.data).any())

    def any_missing(self, *args, ) -> bool:
        """
        """
        return self._func_missing(_np.any, *args, )

    def all_missing(self, *args, ) -> bool:
        """
        """
        return self._func_missing(_np.all, *args, )

    def count_missing(self, *args, ) -> int:
        """
        """
        return self._func_missing(_np.count_nonzero, *args, )

    def _func_missing(self, func, *args, ) -> bool:
        """
        """
        data = self.get_data(*args, )
        return bool(func(_np.isnan(data)))

    def set_data(
        self,
        dates: Dates,
        data: Any | Series,
        variants: VariantsRequestType = None,
    ) -> None:
        """
        """
        dates = self.resolve_periods(dates, )
        is_data_ndarray = isinstance(data, _np.ndarray)
        is_empty_data = (
            data is None
            or (is_data_ndarray and data.size == 0)
        )
        if not dates and is_empty_data:
            return
        data = (
            _reshape_numpy_array(data, )
            if is_data_ndarray else data
        )
        vids = self._resolve_variants(variants, )
        if not self.start:
            self.start = next(iter(dates), None, )
            self.data = self._create_periods_of_missing_values(num_rows=1, )
        pos, add_before, add_after = _get_date_positions(dates, self.start, self.shape[0], )
        self.data = self._create_expanded_data(add_before, add_after)
        if add_before:
            self.start -= add_before
        if hasattr(data, "get_data"):
            data = data.get_data(dates)
        #
        data_variants = _has_variants.iter_variants(data, )
        for c, d in zip(vids, data_variants, ):
            self.data[pos, c] = d
        self.trim()

    def _get_data_and_recreate(
        self,
        *args,
    ) -> _np.ndarray:
        """
        """
        dates, pos, variants, expanded_data = self._resolve_dates_and_positions(*args, )
        data = expanded_data[_np.ix_(pos, variants)]
        num_variants = data.shape[1]
        new = Series(num_variants=num_variants, data_type=self.data_type)
        new.set_data(dates, data)
        return new

    def _resolve_dates_and_positions(
        self,
        dates: Dates,
        variants: VariantsRequestType = None,
    ) -> tuple[Iterable[Period], Iterable[int], Iterable[int], _np.ndarray]:
        """
        """
        dates = self.resolve_periods(dates, )
        variants = self._resolve_variants(variants)
        if not dates:
            dates = ()
            pos = ()
            data = self._create_periods_of_missing_values(num_rows=0, )[:, variants]
            return dates, pos, variants, data
        #
        base_date = self.start or min(dates, )
        pos, add_before, add_after = _get_date_positions(dates, base_date, self.shape[0], )
        data = self._create_expanded_data(add_before, add_after, )
        if not isinstance(pos, Iterable):
            pos = (pos, )
        return dates, pos, variants, data

    def alter_num_variants(
        self,
        new_num: int,
    ) -> Self:
        """
        Alter (expand, shrink) the number of variants in this time series object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )

    def expand_num_variants(
        self,
        new_num: int,
    ) -> None:
        """
        """
        add = new_num - self.num_variants
        if add < 0:
            raise ValueError("Use shrink_num_variants to shrink the number of variants")
        data = _np.tile(self.data[:, (-1, )], (1, add, ))
        self.data = _np.hstack((self.data, data, ), )

    def shrink_num_variants(
        self,
        new_num: int,
    ) -> None:
        """
        """
        remove = self.num_variants - new_num
        if remove < 0:
            raise ValueError("Use expand_num_variants to expand the number of variants")
        self.data = self.data[:, :-remove]

    def get_values(
        self,
        *args,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> list[tuple[Real, ...]] | tuple[Real, ...]:
        """
        """
        data = self.get_data(*args, **kwargs, )
        values = [ tuple(data_column.tolist()) for data_column in data.T ]
        return _has_variants.unpack_singleton(
            values,
            unpack_singleton=unpack_singleton,
        )

    def get_data(self, *args, **kwargs, ) -> _np.ndarray:
        return self.get_data_and_periods(*args, **kwargs, )[0]

    def get_data_and_periods(
        self,
        dates: Iterable[Period] | None | EllipsisType = ...,
        *args,
    ) -> _np.ndarray:
        periods, pos, variants, expanded_values = self._resolve_dates_and_positions(dates, *args, )
        return expanded_values[_np.ix_(pos, variants)], periods

    def get_data_variant(
        self,
        dates: Dates,
        variant: Real | None = None,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant<self.data.shape[1] else 0
        return self.get_data(dates, variant, )

    def get_data_from_until(
        self,
        from_until,
        *args,
    ) -> _np.ndarray:
        """
        """
        _, pos, variants, expanded_data \
            = self._resolve_dates_and_positions(from_until, *args, )
        from_pos, to_pos = pos[0], pos[-1]+1
        return expanded_data[from_pos:to_pos, variants]

    def get_data_variant_from_until(
        self,
        from_until: Iterable[Period],
        variant: int | None = None,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant < self.data.shape[1] else 0
        return self.get_data_from_until(from_until, variant, )

    def extract_variants(
        self,
        variants,
    ) -> None:
        if not isinstance(variants, Iterable):
            variants = (variants, )
        else:
            variants = tuple(c for c in variants)
        self.data = self.data[:, variants]

    def set_start(
        self,
        new_start: Period,
    ) -> Self:
        self.start = new_start
        return self

    def resolve_periods(self, periods, ) -> tuple[Period]:
        """
        """
        if periods is None:
            # periods = []
            periods = ...
        if isinstance(periods, slice) and periods == slice(None, ):
            periods = ...
        if periods is ... and self.start is not None:
            periods = Span(None, None, )
        if periods is ... and self.start is None:
            periods = EmptyRanger()
        if hasattr(periods, "needs_resolve") and periods.needs_resolve:
            periods = periods.resolve(self, )
        return tuple(
            d.resolve(self) if hasattr(d, "needs_resolve") and d.needs_resolve else d
            for d in periods
        )

    def _resolve_variants(
        self,
        variants: VariantsRequestType,
    ) -> Iterable[int]:
        """
        Resolve variant request to an iterable of integers
        """
        if variants is None or variants is Ellipsis:
            variants = slice(None, )
        if isinstance(variants, slice):
            variants = range(*variants.indices(self.num_variants))
        if not isinstance(variants, Iterable):
            variants = (variants, )
        return variants

    def shift(
        self,
        by: ShiftType = -1,
        **kwargs,
    ) -> None:
        r"""
................................................................................

==Shift the time series start date==

Shift the start date of the time series by a number of periods or to a specific
date.

    self.shift(by=-1, )

### Input arguments ###

???+ input "self"
    The current time series object that will be shifted.

???+ input "by"
    The number of periods to shift the observations by. If `by` is a string,
    the observations are manipulated as follows:

    * `"yoy"`: Shift all observations by one year back.

    * `"soy"`: Shift to the start of the year.

    * `"eopy"`: Shift to the end of the previous year.

................................................................................
        """
        if isinstance(by, int):
            self._shift_by_number(by, **kwargs, )
        else:
            method_name = f"_shift_{by}"
            getattr(self, method_name)(**kwargs, )

    def _shift_by_number(self, by: int, **kwargs, ) -> None:
        r"""
        Shift (lag, lead) start date by a number of periods
        """
        if self.start is None:
            return
        self.start -= by

    def _shift_yoy(self, **kwargs, ) -> None:
        r"""
        Shift the start date by one year back
        """
        self._shift_by_number(-self.frequency.value, )

    def _shift_soy(self, **kwargs, ) -> None:
        r"""
        Replace each observation by the start of the year observation
        """
        new_data = self.get_data(
            t.create_soy()
            for t in self.span
        )
        self._replace_data(new_data, )

    def _shift_eopy(self, **kwargs, ) -> Self:
        r"""
        Replace each observation by the end of the previous year observation
        """
        new_data = self.get_data(
            t.create_eopy()
            for t in self.span
        )
        self._replace_data(new_data, )

    def _shift_tty(
        self,
        by: Any | None = None,
        *,
        neutral_value: int | None = None,
        **kwargs,
    ) -> None:
        r"""
        Replace each observation by the start of the year observation except for
        the start-of-year periods which are filled with a neutral value.
        """
        zipped = tuple((t, t.create_tty(), ) for t in self.span)
        neutral_periods = tuple(t for t, tty in zipped if tty is None)
        periods = (t for t, tty in zipped if tty is not None )
        dty_periods = (tty for _, tty in zipped if tty is not None)
        dty_values = self.get_data(dty_periods, )
        self.set_data(periods, dty_values, )
        self.set_data(neutral_periods, neutral_value, )

    def hstack(self, *args, ):
        """
        """
        if not args:
            return self.copy()
        self_args = (self, *args, )
        if all(i.is_empty for i in self_args):
            num_variants = sum(i.num_variants for i in self_args)
            return Series(num_variants=num_variants, )
        encompassing_span, *from_until = _dates.get_encompassing_span(self, *args, )
        new_data = self.get_data_from_until(from_until, )
        add_data = (
            x.get_data_from_until(from_until, )
            if hasattr(x, "get_data")
            else _create_data_variant_from_number(x, encompassing_span, self.data_type)
            for x in args
        )
        new_data = _np.hstack((new_data, *add_data))
        new = Series(num_variants=new_data.shape[1], )
        new.set_data(encompassing_span, new_data, )
        return new

    @_dm.reference(category="homogenizing", )
    def clip(
        self,
        new_start: Period | None,
        new_end: Period | None,
    ) -> None:
        r"""
················································································

==Clip time series to a new start and end period==

    self.clip(new_start, new_end)


### Input arguments ###


???+ input "new_start"
    The new start period for the `self` time series; if `None`, the current
    start period is kept.

???+ input "new_end"
    The new end period for the `self` time series; if `None`, the current
    end period is kept.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        if new_start is None or new_start < self.start:
            new_start = self.start
        if new_end is None or new_end > self.end:
            new_end = self.end
        if new_start == self.start and new_end == self.end:
            return
        self.data = self.get_data_from_until((new_start, new_end, ), )
        self.start = new_start

    def empty(
        self,
        *,
        num_variants: int | None = None
    ) -> None:
        num_variants = num_variants if num_variants is not None else self.num_variants
        self.data = _np.empty((0, num_variants), dtype=self.data_type)

    def redate(
        self,
        new_date: Period,
        old_date: Period | None = None,
    ) -> None:
        """
        """
        if old_data is None:
            self.start = new_date
        else:
            self.start = new_date - (old_date - self.start)

    @_dm.reference(category="manipulation", )
    def replace_where(
        self,
        test: Callable,
        new_value: Real,
    ) -> None:
        """
················································································

==Replace time series values that pass a test==

    self.replace_where(
        test,
        new_value,
    )


### Input arguments ###


???+ input "self"
    Time series whose observations will be tested and those passing the test
    replaced.

???+ input "test"
    A function (or a Callable) that takes a numpy array and returns `True` or
    `False` for each individual value.

???+ input "new_value"
    The value to replace the observations that pass the test.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        self.data[test(self.data)] = new_value
        self.trim()

    def _shallow_copy_data(
        self,
        other: Self,
    ) -> None:
        """
        """
        for n in ("start", "data", "data_type", ):
            setattr(self, n, getattr(other, n, ))

    def __and__(self, other):
        """
        Implement the & operator as hstack
        """
        return self.hstack(other, )

    __or__ = __and__

    @_dm.reference(category="manipulation", )
    def trim(self, ) -> None:
        r"""
................................................................................

==Trim time series data==

Trim leading and trailing missing values from the time series data.

    self.trim()

### Input arguments ###

???+ input "self"
    The time series object to trim.

### Returns ###

This method modifies `self` in place and returns `None`.

................................................................................
        """
        if self.data.size == 0:
            self.reset()
            return self
        num_leading, num_trailing \
            = _get_num_leading_trailing_missing_rows(self.data, )
        if num_trailing == self.data.shape[0]:
            self.reset()
            return self
        if not num_leading and not num_trailing:
            return self
        slice_from = num_leading or None
        slice_to = -num_trailing if num_trailing else None
        self.data = self.data[slice_from:slice_to, ...]
        if slice_from:
            self.start += int(slice_from)
        return self

    def _create_expanded_data(self, add_before, add_after):
        return _np.pad(
            self.data, ((add_before, add_after), (0, 0)),
            mode="constant", constant_values=_np.nan,
        )

    def _check_data_shape(self, data, ):
        if data.shape[1] != self.data.shape[1]:
            raise Exception("Time series data being assigned must preserve the number of variants")

    def __bool__(self, ):
        """
        """
        return self.data.size > 0

    def __neg__(self):
        """
        -self
        """
        output = self.copy()
        output.data = -output.data
        return output

    def __pos__(self):
        """
        +self
        """
        return self.copy()

    def __add__(self, other, ) -> Self | _np.ndarray | Real:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.add, )

    def __radd__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__radd__(other))

    def __mul__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.mul, )

    def __rmul__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__rmul__(other))

    def __sub__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.sub, )

    def __rsub__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rsub__(other))

    def __pow__(self, other):
        """
        self ** other
        """
        # FIXME: 1 ** numpy.nan -> 1 !!!!!
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.pow, )

    def __rpow__(self, other):
        """
        other ** self
        """
        return self.apply(lambda data: data.__rpow__(other))

    def __truediv__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.truediv, )

    def __rtruediv__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rtruediv__(other))

    def __floordiv__(self, other):
        """
        self // other
        """
        return self._binop(other, _op.floordiv, )

    def __rfloordiv__(self, other):
        """
        other // self
        """
        return self.apply(lambda data: data.__rfloordiv__(other))

    def __mod__(self, other):
        """
        self % other
        """
        return self._binop(other, _op.mod, )

    def __rmod__(self, other, ):
        """
        other % self
        """
        return self.apply(lambda data: data.__rmod__(other))

    def __abs__(self, *args, **kwargs, ) -> Self:
        """
        abs(self)
        """
        new = self.copy()
        new.abs(*args, **kwargs, )
        return new

    def abs(self, *args, **kwargs, ) -> None:
        r"""
        """
        self.data = _np.abs(self.data, *args, **kwargs, )

    def __round__(self, *args, **kwargs, ) -> Self:
        r"""
        """
        new = self.copy()
        new.round(*args, **kwargs, )
        return new

    def round(self, *args, **kwargs, ) -> None:
        self.data = _np.round(self.data, *args, **kwargs, )

    for n in ["gt", "lt", "ge", "le", "eq", "ne", ]:
        exec(f"def __{n}__(self, other): return self._binop(other, _op.{n}, )", )

    @_dm.reference(category="comparison", )
    def same_as(self, other: Self, ) -> bool:
        r"""
................................................................................

==Check if two time series are the same==

    self.same_as(other)


### Input arguments ###

???+ input "self"
    The current time series object.

???+ input "other"
    The time series object to compare with the current time series object.


### Returns ###

???+ returns "bool"
    `True` if the two time series are the same by comparing their start dates,
    numerical values, and descriptions; `False` otherwise.

................................................................................
        """
        if not isinstance(other, type(self)):
            return False
        if self.frequency != other.frequency:
            return False
        if self.start != other.start:
            return False
        if not _np.array_equal(self.data, other.data, equal_nan=True, ):
            return False
        if self.get_description() != other.get_description():
            return False
        return True

    def apply(self, func, *args, **kwargs, ):
        new_data = func(self.data, *args, **kwargs, )
        axis = kwargs.get("axis", None, )
        if new_data.shape == self.data.shape:
            new = self.copy()
            new._replace_data(new_data, )
            return new
        elif (axis is None or axis == 1) and new_data.shape == (self.data.shape[0], ):
            new = self.copy()
            new.data = new_data.reshape(self.data.shape[0], 1, )
            new.trim()
            return new
        elif axis is None and new_data.shape == ():
            return new_data
        elif axis == 0 and new_data.shape == (self.data.shape[1], ):
            return new_data
        else:
            raise _wrongdoings.IrisPieError(
                "Function applied on a time series resulted"
                " in a data array with an unexpected shape"
            )

    def _binop(self, other, func, new=None, ):
        if not isinstance(other, type(self)):
            return self.apply(lambda data: func(data, other))
        # FIXME: empty encompassing span
        _, *from_until = _dates.get_encompassing_span(self, other)
        self_data = self.get_data_from_until(from_until, )
        other_data = other.get_data_from_until(from_until, )
        new_data = func(self_data, other_data)
        new = Series(num_variants=new_data.shape[1], ) if new is None else new
        new._replace_start_and_values(from_until[0], new_data, )
        return new


    def _replace_data(
        self,
        new_values,
    ) -> None:
        """
        """
        self.data = new_values
        self.trim()

    def _replace_start_and_values(
        self,
        new_start,
        new_values,
    ) -> None:
        """
        """
        self.start = new_start
        self._replace_data(new_values, )

    def iter_dates_values(self, unpack_singleton=True, ):
        """
        Default iterator is line by line, yielding a tuple of (date, values)
        """
        def _unpack_singleton_data_row(data_list: list[Real], ):
            return _has_variants.unpack_singleton(data_list, True, )
        def _keep_data_row(data_list: list[Real], ):
            return data_list
        data_row_func = (
            _unpack_singleton_data_row
            if self.is_singleton and unpack_singleton
            else _keep_data_row
        )
        for date, data_row in zip(self.span, self.data, ):
            yield date, data_row_func(data_row.tolist(), )

    def iter_variants(self, ) -> Iterator[Self]:
        """
        """
        for data in self.iter_data_variants_from_until(..., ):
            new = self.copy()
            new._replace_data(data, )
            yield new

    def iter_own_data_variants_from_until(self, from_until, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        if from_until == ...:
            data_from_until = self.data
        else:
            data_from_until = self.get_data_from_until(from_until, )
        return iter(data_from_until.T, )

    def iter_data_variants_from_until(self, from_until, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        return _iterators.exhaust_then_last(self.iter_own_data_variants_from_until(from_until, ), )

    def logistic(self, ) -> Self:
        """
        """
        self.data = 1 / (1 + _np.exp(-self.data))

    #]


def _get_num_leading_trailing_missing_rows(data: _np.ndarray, ):
    r"""
    """
    #[
    # Boolean index of rows with at least one observation
    boolex_observations = ~_np.all(_np.isnan(data, ), axis=1, )
    if _np.all(boolex_observations, ):
        # All rows have at least one observation
        num_leading = 0
        num_trailing = 0
    elif not _np.any(boolex_observations, ):
        # No row has any observation
        num_leading = 0
        num_trailing = data.shape[0]
    else:
        # Some rows have observations
        num_leading = _np.argmax(boolex_observations)
        num_trailing = _np.argmax(boolex_observations[::-1])
    return num_leading, num_trailing
    #]


def hstack(first, *args) -> Self:
    return first.hstack(*args)


def _create_data_variant_from_number(
    number: Real,
    span: Span,
    data_type: type,
) -> _np.ndarray:
    return _np.full((len(span), 1), number, dtype=data_type)


def _from_periods_and_values(
    self,
    periods: Iterable[Period] | str,
    values: _np.ndarray | Iterable,
    frequency: Frequency | None = None,
    **kwargs,
) -> None:
    r"""
    """
    #[
    # dates = _dates.ensure_period_tuple(dates, frequency=frequency, )
    self.set_data(periods, values, )
    #]


def _from_periods_and_func(
    self,
    periods: Iterable[Period] | str,
    func: Callable,
    frequency: Frequency | None = None,
    **kwargs,
) -> Self:
    """
    Create a new time series from dates and a function
    """
    #[
    # dates = _dates.ensure_period_tuple(dates, frequency=frequency, )
    data = [
        [func() for j in range(self.num_variants)]
        for i in range(len(periods))
    ]
    data = _np.array(data, dtype=self.data_type)
    self.set_data(periods, data)
    #]


def _from_start_and_values(
    self,
    start: Period,
    values: _np.ndarray | Iterable,
    frequency: Frequency | None = None,
    trim: bool = True,
    **kwargs,
) -> None:
    """
    """
    #[
    # start = _dates.ensure_period_tuple(start, frequency=frequency, )[0]
    self.start = start
    if isinstance(values, _np.ndarray):
        values = _reshape_numpy_array(values, )
    else:
        values = _has_variants.iter_variants(values, )
        values = _np.column_stack([
            v for v, _ in zip(values, range(self.num_variants), )
        ])
    values = values.astype(self.data_type, )
    self.data = values
    #
    if trim:
        self.trim()
    #]


def _reshape_numpy_array(values: _np.ndarray, ) -> _np.ndarray:
    r"""
    Reshape numpy array to a 2D array
    """
    #[
    if values.ndim == 2:
        return values
    if values.ndim < 2:
        return values.reshape(-1, 1, )
    if values.ndim > 2:
        return values.reshape(values.shape[0], -1, )
    #]


def _invalid_constructor(
    self,
    **kwargs,
) -> NoReturn:
    raise _wrongdoings.IrisPieError("Invalid Series object constructor")


_SERIES_POPULATOR = {
    (False, False, False, False): lambda self, **kwargs: None,
    (True, False, True, False): _from_start_and_values,
    (False, True, True, False): _from_periods_and_values,
    (False, True, False, True): _from_periods_and_func,
}


FUNCTIONAL_FORMS = (
    "shift",
    "redate",
)




"""
Mixin for plan registers
"""


#[
from __future__ import annotations

from .. import wrongdoings as _wrongdoings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Callable, NoReturn, Any, Iterable, EllipsisType
#]


class Mixin:
    """
    """
    #[

    def _initialize_registers(
        self,
        plannable,
        default_value: Callable,
    ) -> None:
        """
        """
        for n in self._registers:
            can_be_name = f"can_be_{n}"
            register = {
                n: default_value(n, )
                for n in getattr(plannable, can_be_name, )
            } if hasattr(plannable, can_be_name, ) else {}
            setattr(self, can_be_name, tuple(register.keys()))
            setattr(self, f"_{n}_register", register)

    def get_register_by_name(self, name: str, /, ) -> dict[str, Any]:
        """
        """
        full_name = f"_{name}_register"
        return getattr(self, full_name, ) if hasattr(self, full_name, ) else None

    @staticmethod
    def _resolve_register_names(
        register: dict | None,
        names: Iterable[str] | str | EllipsisType,
    ) -> tuple[str] | NoReturn:
        """
        """
        keys = tuple(register.keys()) if register else ()
        if names is Ellipsis:
            names = keys
        elif isinstance(names, str):
            names = (names, )
        else:
            names = tuple(names)
        _validate_register_names(register, names, )
        return names

    #]


def _validate_register_names(
    register: dict | None,
    names: Iterable[str],
) -> None | NoReturn:
    """
    """
    #[
    keys = tuple(register.keys()) if register else ()
    invalid = tuple(n for n in names if n not in keys)
    if invalid:
        message = (f"These names are not valid in the register:", ) + invalid
        raise _wrongdoings.IrisPieCritical(message, )
    #]




"""
Facades for steady-state solvers
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp
import neqs as _ne

from .evaluators import SteadyEvaluator

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    SolverType = Literal["neqs_levenberg", "neqs_firefly", "scipy_root", ]

#]


#===============================================================================


_DEFAULT_SCIPY_ROOT_SETTINGS = {
    "method": "lm",
    "tol": None,
    "norm_order": 2,
}


def scipy_root(
    steady_evaluator: SteadyEquator,
    maybelog_init_guess: _np.ndarray,
    solver_settings: dict[str, Any] | None,
) -> tuple[bool, _np.ndarray, Any, ]:
    """
    """
    pass_solver_settings = dict(solver_settings, )
    del pass_solver_settings["norm_order"]
    root_final = _sp.optimize.root(
        steady_evaluator.eval,
        maybelog_init_guess,
        jac=True,
        **pass_solver_settings,
    )
    #
    steady_evaluator.iter_printer.print_footer()
    maybelog_final_guess = root_final.x
    func_norm = _sp.linalg.norm(root_final.fun, 2, )
    success = root_final.success and func_norm < solver_settings["tol"]
    # TODO: create exit status based on root_final
    exit_status = None
    return maybelog_final_guess, success, exit_status,


def create_solver_settings_for_scipy_root(
    user_solver_settings: dict[str, Any] | None,
    self_equality_tolerance: float,
) -> dict[str, Any]:
    r"""
    """
    #[
    solver_settings = dict(_DEFAULT_SCIPY_ROOT_SETTINGS, )
    solver_settings["tol"] = self_equality_tolerance
    if user_solver_settings:
        solver_settings.update(user_solver_settings, )
    return solver_settings
    #]


#===============================================================================


_DEFAULT_NEQS_LEVENBERG_SETTINGS = {
    "func_tolerance": None,
    "step_tolerance": float("inf"),
    "norm_order": float("inf"),
}


def neqs_levenberg(
    steady_evaluator: SteadyEquator,
    maybelog_init_guess: _np.ndarray,
    solver_settings: dict[str, Any],
) -> tuple[bool, _np.ndarray, Any, ]:
    """
    """
    maybelog_final_guess, exit_status = _ne.levenberg(
        eval_func=steady_evaluator.eval_func,
        eval_jacob=steady_evaluator.eval_jacob,
        init_guess=maybelog_init_guess,
        iter_printer=steady_evaluator.iter_printer,
        **solver_settings,
    )
    success = exit_status.is_success
    return maybelog_final_guess, success, exit_status,


def create_solver_settings_for_neqs_levenberg(
    user_solver_settings: dict[str, Any] | None,
    self_equality_tolerance: float,
) -> dict[str, Any]:
    """
    Create a dictionary of settings for the NEQS Levenberg solver.
    """
    #[
    solver_settings = dict(_DEFAULT_NEQS_LEVENBERG_SETTINGS, )
    solver_settings["func_tolerance"] = self_equality_tolerance
    if user_solver_settings:
        solver_settings.update(user_solver_settings, )
    return solver_settings
    #]


#===============================================================================


# _DEFAULT_NEQS_FIREFLY_SETTINGS = {
# }
# 
# 
# def neqs_firefly(
#     steady_evaluator: SteadyEquator,
#     maybelog_init_guess: _np.ndarray,
#     *,
#     solver_settings: dict[str, Any] | None = None,
# ) -> tuple[bool, _np.ndarray, Any, ]:
#     """
#     """
#     solver_settings = solver_settings or {}
#     solver_settings = _DEFAULT_NEQS_FIREFLY_SETTINGS | solver_settings
#     #
#     maybelog_final_guess = _ne.firefly(
#         eval_func=steady_evaluator.eval_func,
#         init_guess=maybelog_init_guess,
#         iter_printer=steady_evaluator.iter_printer,
#         solver_settings=solver_settings,
#     )
#     #
#     success = True
#     exit_status = None
#     return maybelog_final_guess, success, exit_status,





r"""
"""


#[

from __future__ import annotations

from typing import Literal
import documark as _dm
import numpy as _np

from .. import dates as _dates
from . import _broadcasts as _bc

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .main import Series

#]


__all__ = ()


LayMethod = Literal["by_span", "by_observation"]


def mixin(klass: type, ) -> type:
    r"""
    Inlay the steady state methods in the class
    """
    #[
    klass.overlay_by_span = overlay_by_span
    klass.overlay_by_observation = overlay_by_observation
    klass.underlay_by_span = underlay_by_span
    klass.underlay_by_observation = underlay_by_observation
    return klass
    #]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Series class
#-------------------------------------------------------------------------------


@_dm.no_reference
def overlay_by_span(
    self: Series,
    other: Series,
) -> None:
    r"""
................................................................................

==Overlay the current series with another series by span==

Overlay the values of another time series onto the current time series on the
entire span of the other time series, i.e. from the start to the end period
regardless of missing in-sample values.

    self.overlay_by_span(
        other,
    )

### Input arguments ###

???+ input "self"
The current time series object.

???+ input "other"
The time series object whose values will be overlaid onto the current time
series.

???+ input "method"
The method to use for overlaying the values. The default (and currently the
only available) method is `"by_span"`.

### Returns ###

This method modifies `self` in place and returns `None`.

### Details ###

???+ abstract "Algorithm"

The resulting time series is determined the following way:

1. The span of the resulting series starts at the earliest start period of the two
series and ends at the latest end period of the two series.

2. The observations from the `self` (current) time series used to fill the
resulting time span.

3. Within the span of the `other` time series (from the first available
observation to the last available observation), the observations from this
`other` time series are superimposed on the resulting time series, including any
in-sample missing observations.

................................................................................
    """
    other_copy = other.copy()
    _bc.broadcast_variants_when_needed(self, other_copy, )
    self.set_data(other_copy.span, other_copy.data, )
    self.trim()


@_dm.no_reference
def overlay_by_observation(
    self,
    other,
) -> None:
    r"""
................................................................................

==Overlay the current series with another series by observation==

Overlay the values of another time series onto the current time series
observation by observation, only where the other time series has valid
(non-missing) observations.

    self.overlay_by_observation(
        other,
    )

### Input arguments ###

???+ input "self"
    The current time series object.

???+ input "other"
    The time series object whose values will be overlaid onto the current time
    series.

### Returns ###

This method modifies `self` in place and returns `None`.

### Details ###

???+ abstract "Algorithm"

    The resulting time series is determined the following way:

    1. The span of the resulting series starts at the earliest start period of the two
    series and ends at the latest end period of the two series.

    2. The observations from the `self` (current) time series are used to fill the
    resulting time span.

    3. For each period where the `other` time series has a valid (non-missing)
    observation, that observation is superimposed on the resulting time series,
    replacing the value from `self` at that period.

................................................................................
    """
    # Handle empty series cases
    if self.is_empty and other.is_empty:
        return
    #
    if self.is_empty:
        self.start = other.start
        self.data = _np.array(other.data)
        return
    #
    if other.is_empty:
        return
    #
    # Broadcast variants if needed
    _bc.broadcast_variants_when_needed(self, other)
    #
    # Get encompassing span and from_until tuple for data extraction
    encompassing_span, *from_until = _dates.get_encompassing_span(self, other)
    #
    # Get data for both series over the encompassing span
    self_data = self.get_data_from_until(from_until)
    other_data = other.get_data_from_until(from_until)
    #
    # Create boolean indices for non-NaN values
    other_valid = ~_np.isnan(other_data)
    #
    # For overlay: use other's values where other has valid data
    result_data = _np.array(self_data)
    result_data[other_valid] = other_data[other_valid]
    #
    # Set the result data directly
    self.start = encompassing_span.start
    self.data = result_data


@_dm.no_reference
def underlay_by_span(
    self,
    other,
) -> None:
    r"""
................................................................................

==Underlay the current series with another series by span==

Underlay the values of another time series beneath the current time series on
the entire span of the other time series, i.e. from the start to the end period
regardless of missing in-sample values.

    self.underlay_by_span(
        other,
    )

### Input arguments ###

???+ input "self"
The current time series object.

???+ input "other"
The time series object whose values will be underlaid beneath the current
time series.

???+ input "method"
The method to use for underlaying the values. The default (and currently the
only available) method is `"by_span"`.

### Returns ###

This method modifies `self` in place and returns `None`.

### Details ###

???+ abstract "Algorithm"

The resulting time series is determined the following way:

1. The span of the resulting series starts at the earliest start period of the two
series and ends at the latest end period of the two series.

2. The observations from the `other` time series used to fill the
resulting time span.

3. Within the span of the `self` time series (from the first available
observation to the last available observation), the observations from this
`self` time series are superimposed on the resulting time series, including any
in-sample missing observations.

................................................................................
    """
    new_self = other.copy()
    new_self.overlay_by_span(self, )
    self._shallow_copy_data(new_self, )


@_dm.no_reference
def underlay_by_observation(
    self,
    other,
) -> None:
    r"""
................................................................................

==Underlay the current series with another series by observation==

Underlay the values of another time series beneath the current time series
observation by observation, only where the current time series has missing
observations.

    self.underlay_by_observation(
        other,
    )

### Input arguments ###

???+ input "self"
    The current time series object.

???+ input "other"
    The time series object whose values will be underlaid beneath the current
    time series.

### Returns ###

This method modifies `self` in place and returns `None`.

### Details ###

???+ abstract "Algorithm"

    The resulting time series is determined the following way:

    1. The span of the resulting series starts at the earliest start period of the two
    series and ends at the latest end period of the two series.

    2. The observations from the `other` time series are used to fill the
    resulting time span.

    3. For each period where the `self` time series has a valid (non-missing)
    observation, that observation is superimposed on the resulting time series,
    replacing the value from `other` at that period.

................................................................................
    """
    new_self = other.copy()
    new_self.overlay_by_observation(self)
    self._shallow_copy_data(new_self)


#-------------------------------------------------------------------------------


FUNCTIONAL_FORMS = (
    "overlay_by_span",
    "overlay_by_observation",
    "underlay_by_span",
    "underlay_by_observation",
)




"""
Printers for iterative solvers
"""


#[

from __future__ import annotations

import os as _os
import numpy as _np
import scipy as _sp
import functools as _ft

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any
    from numbers import Number
    from collections.abc import Iterable

#]


_NONE_SYMBOL = "•"
_CONTINUATION_SYMBOL = "⋯"
_LEN_CONTINUATION_SYMBOL = len(_CONTINUATION_SYMBOL)
_DEFAULT_PRINT_EVERY = 5
_DEFAULT_NORM_ORDER = 2
_ROUND_FOR_PRINT = 99

class EmptyIterPrinter:
    """
    Iterations printer for steady equator
    """
    #[

    _COLUMN_CLASSES = ()

    _HEADER_DIVIDER_CHAR = "-"
    _COLUMN_DIVIDER = " "*3

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        """
        # self._populate_equations(equations, )
        # self._populate_quantities(qids, qid_to_logly, qid_to_name, )
        self.reset()
        self.custom_header = kwargs.get("custom_header", None, )
        self._print_every = kwargs.get("print_every", _DEFAULT_PRINT_EVERY, )
        self._columns = tuple(i(**kwargs, ) for i in self._COLUMN_CLASSES)

    def reset(self, /, ) -> None:
        """
        """
        self._count = 0
        self.custom_header = None
        self._last_iter_string = None
        self.num_equations = None
        self.num_quantities = None

    def next(self, x, f, jacobian_calculated: bool, ) -> None:
        """
        Handle next function evaluation
        """
        if self._count == 0:
            self._first_iteration(x, f, jacobian_calculated, )
        for i in self._columns:
            i.next(self._count, x, f, jacobian_calculated, )
        self._last_iter_string = self._compose_iter_string()
        if self._count % self._print_every == 0:
            _print_to_width(self._last_iter_string)
        self._count += 1

    def _first_iteration(self, x, f, jacobian_calculated, ) -> None:
        self.num_equations = len(f)
        self.num_quantities = len(x)
        self.print_header()

    def print_header(self, /, ) -> None:
        """
        Print header for fuction evaluations
        """
        custom_header = str(self.custom_header or "")
        dimension_header = f"[Dimension {self.num_equations}×{self.num_quantities}]"
        header = self._COLUMN_DIVIDER.join(i.get_header_string() for i in self._columns)
        len_header = len(header)
        upper_line = self._HEADER_DIVIDER_CHAR + f"{custom_header}{dimension_header}" + self._HEADER_DIVIDER_CHAR * len_header
        upper_line = upper_line[:len_header]
        self._divider_line = self._HEADER_DIVIDER_CHAR * len_header
        _print_to_width("", upper_line, header, self._divider_line, )

    def conclude(self, /, ) -> None:
        """
        """
        self.print_footer()

    def print_footer(self, /, ) -> None:
        """
        Print footer for iterations
        """
        _print_to_width(self._divider_line, self._last_iter_string, self._divider_line, )

    def _compose_iter_string(self, /, ) -> str:
        return self._COLUMN_DIVIDER.join(i.get_iter_string() for i in self._columns)

    #]


class CounterColumn:
    #[

    _HEADER_SYMBOL = "ƒ-count"
    _WIDTH = 8

    def __init__(self, **kwargs, ) -> None:
        self._count = None

    def reset(self, /, ) -> None:
        self._count = None

    def next(self, count: int, x, f, jacobian_calculated: bool, /, ) -> None:
        self._count = count

    def get_header_string(self, /, ) -> str:
        return f"{self._HEADER_SYMBOL:>8}"

    def get_iter_string(self, /, ) -> str:
        return f"{self._count:{self._WIDTH}g}"

    #]


class FuncNormColumn:
    #[

    _HEADER_SYMBOL = "‖ƒ‖"
    _WIDTH_DECIMALS = 5
    _WIDTH_EXPONENT = 4
    _WIDTH = 2 + _WIDTH_DECIMALS + _WIDTH_EXPONENT

    def __init__(self, **kwargs, ) -> None:
        self._norm = None
        self._eval_norm = _ft.partial(_sp.linalg.norm, ord=kwargs.get("norm_order", _DEFAULT_NORM_ORDER), )

    def reset(self, /, ) -> None:
        self._norm = None

    def next(self, count: int, x, f, jacobian_calculated: bool, /, ) -> None:
        self._norm = self._eval_norm(f)

    def get_header_string(self, /, ) -> str:
        return f"{self._HEADER_SYMBOL:>{self._WIDTH}}"

    def get_iter_string(self, /, ) -> str:
        value_to_print = round(self._norm, _ROUND_FOR_PRINT, )
        return f"{value_to_print:.{self._WIDTH_DECIMALS}e}"

    #]


class JacobRecalculatedColumn:
    #[

    _HEADER_SYMBOL = "∇ƒ"
    _STATUS_SYMBOL = {True: "√", False: "×"}
    _WIDTH = 2

    def __init__(self, **kwargs, ) -> None:
        self._status = None

    def reset(self, /, ) -> None:
        self._status = None

    def next(self, count: int, x, f, jacobian_calculated: bool, /, ) -> None:
        self._status = jacobian_calculated

    def get_header_string(self, /, ) -> str:
        return f"{self._HEADER_SYMBOL:>{self._WIDTH}}"

    def get_iter_string(self, /, ) -> str:
        return f"{self._STATUS_SYMBOL[self._status]:>{self._WIDTH}}"

    #]


class _WorstColumn:
    #[

    _HEADER_SYMBOL = ...
    _NAME_SYMBOL = ...
    _WIDTH_NAME = ...

    _JOIN_SYMBOL = " "
    _WIDTH_JOIN_SYMBOL = len(_JOIN_SYMBOL)
    _WIDTH_DECIMALS = 5
    _WIDTH_EXPONENT = 4
    _WIDTH_NUMERIC = 2 + _WIDTH_DECIMALS + _WIDTH_EXPONENT

    def __init__(self, **kwargs, ) -> None:
        self._populate_names(**kwargs, )
        self._worst_value = None
        self._worst_name = None
        self._prev = None

    @property
    def _WIDTH(self, /, ) -> int:
        return self._WIDTH_NUMERIC + self._WIDTH_JOIN_SYMBOL + self._WIDTH_NAME

    def _populate_names(self, **kwargs, ) -> None:
        ...

    def _get_name(self, index, ) -> str:
        return (
            self._names[index % len(self._names)] if self._names
            else f"{self._NAME_SYMBOL}[{index}]"
        )

    def reset(self, /, ) -> None:
        self._worst_values = None
        self._worst_name = None
        self._prev = None

    def next(self, count: int, x, f, jacobian_calculated: bool, ) -> None:
        ...

    def get_header_string(self, /, ) -> str:
        header_numeric = f"{self._HEADER_SYMBOL:>{self._WIDTH_NUMERIC}}"
        return f"{header_numeric:<{self._WIDTH}}"

    def get_iter_string(self, /, ) -> str:
        return (
            self._get_proper_iter_string()
            if self._worst_value is not None
            else self._get_none_iter_string()
        )

    def _get_proper_iter_string(self, /, ) -> str:
        value_to_print = round(self._worst_value, _ROUND_FOR_PRINT, )
        return (
            f"{value_to_print:.{self._WIDTH_DECIMALS}e}"
            f"{self._JOIN_SYMBOL}"
            f"{self._worst_name:>{self._WIDTH_NAME}}"
        )

    def _get_none_iter_string(self, /, ) -> str:
        return (
            f"{_NONE_SYMBOL:>{self._WIDTH_NUMERIC}}"
            f"{' ':>{self._WIDTH_JOIN_SYMBOL}}"
            f"{' ':>{self._WIDTH_NAME}}"
        )

    #]


class WorstDiffXColumn(_WorstColumn, ):
    #[

    _HEADER_SYMBOL = "max|∆x|"
    _WIDTH_NAME = 10

    def _populate_names(self, **kwargs, ) -> None:
        self._names = tuple(
            _clip_string_exactly(i, self._WIDTH_NAME, )
            for i in kwargs["quantity_strings"]
        ) if "quantity_strings" in kwargs else None

    def next(self, count: int, x, f, jacobian_calculated: bool, ) -> None:
        self._worst_value = None
        self._worst_name = None
        if self._prev is not None:
            diff = abs(x - self._prev)
            index = _np.argmax(diff)
            self._worst_value = diff[index]
            self._worst_name = self._get_name(index, )
        self._prev = x

    #]


class WorstFuncColumn(_WorstColumn, ):
    #[

    _HEADER_SYMBOL = "max|ƒ|"
    _WIDTH_NAME = 25

    def _populate_names(self, **kwargs, ) -> None:
        self._names = tuple(
            _clip_string_exactly(i, self._WIDTH_NAME, )
            for i in kwargs["equation_strings"]
        ) if "equation_strings" in kwargs else None

    def next(self, count: int, x, f, jacobian_calculated: bool, ) -> None:
        abs_f = _np.abs(f)
        index = _np.argmax(abs_f)
        self._worst_value = abs_f[index]
        self._worst_name = self._get_name(index, )

    #]


def _clip_string_exactly(
    full_string: str,
    max_length: int,
    /,
) -> str:
    """
    Clip string to exactly to max length
    """
    return (
        full_string[:max_length-_LEN_CONTINUATION_SYMBOL] + _CONTINUATION_SYMBOL
        if len(full_string) > max_length
        else f"{full_string:<{max_length}}"
    )


def _print_to_width(*args, ) -> None:
    """
    """
    try:
        width = _os.get_terminal_size().columns
    except:
        width = None
    for text in args:
        if width is not None and len(text) > width:
            text = text[:width-1] + "⋯"
        print(text)


class IterPrinter(EmptyIterPrinter, ):
    """
    """
    #[

    _COLUMN_CLASSES = (
        CounterColumn,
        FuncNormColumn,
        JacobRecalculatedColumn,
        WorstDiffXColumn,
        WorstFuncColumn,
    )

    #]




"""
Dynamic function makers
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Any, Callable, )
import copy as _cp

from .aldi import adaptations as _adaptations
#]


def make_function(
    func_name: str,
    args: Iterable[str],
    expression: str,
    context: dict[str, Any] | None,
    /,
) -> tuple[Callable, str, dict[str, Any]]:
    """
    Create lambda from the list of args and an expression
    """
    #[
    globals_ = _prepare_globals(context, )
    args_string = ', '.join(args, )
    func_str = f"def {func_name}({args_string}): return {str(expression)}"
    exec(func_str, globals_, )
    return globals_[func_name], func_str, globals_,
    #]


def _prepare_globals(
    context: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """
    """
    #[
    globals_ = (context or {}) | {"__builtins__": {}}
    globals_ = _adaptations.add_function_adaptations_to_context(globals_, )
    return globals_
    #]




"""
Plotly interface to time series objects
"""


#[

from __future__ import annotations

import os as _os
import json as _js
import copy as _cp
import plotly.graph_objects as _pg
import plotly.io as _pi
import itertools as _it
import warnings as _wa
import datetime as _dt

from ..dates import Period, PositionType
from .. import dates as _dates
from .. import ez_plotly as _ez_plotly

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Sequence, Iterable, Any, Callable, Literal
    from numbers import Real
    from types import EllipsisType
    from . import Series
    from ..ez_plotly import PlotlyDateAxisModeType

#]


_PLOTLY_STYLES_FOLDER = _os.path.join(_os.path.dirname(__file__), "plotly_styles")
_PLOTLY_STYLES = {
    "layouts": {},
}
with open(_os.path.join(_PLOTLY_STYLES_FOLDER, "plain_layout.json", ), "rt", ) as fid:
    _PLOTLY_STYLES["layouts"]["plain"] = _js.load(fid, )


def _line_plot(color: str, x, y, **settings) -> _pg.Scatter:
    """
    """
    settings = {"mode": "lines+markers", } | settings
    return _pg.Scatter(line_color=color, x=x, y=y, **settings, )


def _bar_plot(color: str, **settings) -> _pg.Bar:
    """
    """
    return _pg.Bar(marker_color=color, **settings, )


_PLOTLY_TRACES_CONSTRUCTOR = {
    "line": _line_plot,
    "bar": _bar_plot,
    "bar_relative": _bar_plot,
    "bar_group": _bar_plot,
    "bar_stack": _bar_plot,
    "bar_overlay": _bar_plot,
}


_BARMODE = {
    "bar": "group",
    "bar_relative": "relative",
    "bar_group": "group",
    "bar_stack": "stack",
    "bar_overlay": "overlay",
}


#_COLOR_ORDER = [
#    "rgb(  0,   114,   189)",
#    "rgb(217,    83,    25)",
#    "rgb(237,   177,    32)",
#    "rgb(126,    47,   142)",
#    "rgb(119,   172,    48)",
#    "rgb( 77,   190,   238)",
#    "rgb(162,    20,    47)",
#]


_COLOR_ORDER = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]


class Inlay:
    """
    """
    #[

    def plot(
        self,
        *,
        span: Iterable[Period] | EllipsisType = ...,
        figure: _pg.Figure | None = None,
        figure_title: str | None = None,
        subplot_title: str | None = None,
        legend: Iterable[str] | None = None,
        show_figure: bool = True,
        show_legend: bool | None = None,
        include_in_legend: bool | Iterable[bool] = False,
        renderer: str | None = None,
        subplot: tuple[int, int] | int | None = None,
        xline = None,
        vline = None,
        type = None,
        chart_type: Literal["line", "bar_stack", "bar_group", ] = "line",
        highlight: Iterable[Period] | None = None,
        bar_norm: str | None = None,
        date_axis_mode: PlotlyDateAxisModeType = "period",
        date_format_style: str | dict[int, str] | None = None,
        #
        update_layout: dict[str, Any] | None = None,
        update_traces: tuple(dict[str, Any], ) | dict | None = None,
        #
        freeze_span: bool = False,
        reverse_plot_order: bool = False,
        #
        round: int | None = None,
        round_to: int | None = None,
        #
        return_info: bool = False,
    ) -> dict[str, Any] | None:
        """
        """
        if type is not None:
            _wa.warn("Use 'chart_type' instead of the deprecated 'type'", )
            chart_type = type if chart_type is None else chart_type

        if round is not None:
            _wa.warn("Use 'round_to' instead of the deprecated 'round'", )
            round_to = round if round_to is None else round_to

        span = tuple(self.resolve_periods(span, ))
        from_until = (span[0], span[-1], )
        frequency = span[0].frequency
        date_format = _ez_plotly.resolve_date_format(date_format_style, frequency, )

        if figure is None:
            figure = _pg.Figure()

        # Subplot resolution
        row_column, index = _ez_plotly.resolve_subplot_reference(figure, subplot, )

        if show_legend is None:
            show_legend = legend is not None

        color_cycle = _it.cycle(_COLOR_ORDER)

        if isinstance(update_traces, dict):
            update_traces = (update_traces, )

        update_traces_cycle = (
            _it.cycle(update_traces) if update_traces
            else _it.repeat({}, )
        )

        transform = _create_transform_function(round_to=round_to, )

        if legend is None:
            legend = _it.repeat(None, )

        traces_constructor = _PLOTLY_TRACES_CONSTRUCTOR[chart_type]

        xaxis_type = from_until[0].plotly_xaxis_type if from_until else None

        traces_periods = tuple(
            i.to_plotly_date(mode=date_axis_mode, )
            for i in _dates.periods_from_until(*from_until, )
        )

        traces_iterable = _iter_traces(
            self,
            from_until,
            traces_periods,
            traces_constructor,
            transform,
            legend,
            color_cycle,
            update_traces_cycle,
            include_in_legend=include_in_legend,
            xhoverformat=date_format,
        )

        if reverse_plot_order:
            traces_iterable = reversed(list(traces_iterable))

        out_traces = ()
        traces_offset = len(tuple(figure.select_traces()))
        for tid, traces_v, in enumerate(traces_iterable, start=traces_offset, ):
            custom_data = (tid, )
            traces_v.update(customdata=custom_data, )
            figure.add_trace(traces_v, **row_column, )
            out_traces += tuple(figure.select_traces({"customdata": custom_data}, ))

        # REFACTOR
        xaxis = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"]["xaxis"])
        yaxis = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"]["yaxis"])
        layout = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"])
        del layout["xaxis"]
        del layout["yaxis"]

        layout["showlegend"] = show_legend

        bar_mode = _BARMODE.get(chart_type, figure.layout["barmode"], )
        layout["barmode"] = bar_mode
        layout["barnorm"] = bar_norm

        xaxis["tickformat"] = date_format
        xaxis["ticklabelmode"] = date_axis_mode

        if xaxis_type is not None:
            xaxis["type" ] = xaxis_type

        figure.update_xaxes(xaxis, **row_column, )
        figure.update_yaxes(yaxis, **row_column, )
        figure.update_layout(layout or {}, )

        if freeze_span:
            _ez_plotly.freeze_span(figure, span, subplot, )

        if figure_title is not None:
            figure.update_layout({"title.text": figure_title, }, )

        if update_layout is not None:
            figure.update_layout(update_layout, )

        if subplot_title:
            _ez_plotly.update_subplot_title(figure, index, subplot_title, )

        if xline is not None and vline is None:
            vline = xline

        if vline is not None:
            _ez_plotly.add_vline(figure, vline, )

        if highlight is not None:
            _ez_plotly.add_highlight(figure, highlight, )

        if show_figure:
            figure.show(renderer=renderer, )

        if not return_info:
            return

        out_info = {
            "figure": figure,
            "traces": out_traces,
        }
        return out_info

    #]


# def _get_traces_dates(series: Series, /, ) -> tuple[tuple[Period, ...], str]:
#     """
#     """
#     #[
#     if not series.span:
#         return (), None,
#     dates = tuple(i.to_plotly_date() for i in series.span)
#     xaxis_type = next(iter(series.span)).plotly_xaxis_type
#     return dates, xaxis_type,
#    #]


def _iter_traces(
    series: Series,
    from_until: tuple[Period, Period],
    traces_period: tuple[str, ...],
    traces_constructor: Callable,
    transform: Callable,
    legends: Iterable[str] | None,
    colors: Iterable[str],
    custom_updates: Iterable[dict],
    include_in_legend: Iterable[bool],
    **kwargs,
) -> Any:
    """
    """
    #[
    if isinstance(include_in_legend, bool):
        include_in_legend = _it.repeat(include_in_legend, )
    #
    zipped = zip(
        series.iter_own_data_variants_from_until(from_until=from_until, ),
        legends,
        colors,
        custom_updates,
        include_in_legend,
    )
    for data_v, legend_v, color_v, update_v, include_in_legend_v in zipped:
        traces = traces_constructor(
            color=color_v,
            x=traces_period,
            y=tuple(transform(i) for i in data_v),
            name=legend_v,
            showlegend=include_in_legend_v,
            **kwargs,
        )
        traces.update(update_v, )
        yield traces
    #]


def _create_transform_function(
    round_to: int | None = None,
) -> Callable[[Real], Real]:
    """
    """
    #[
    def no_transform(x: Real, ) -> Real:
        return x

    def decorate_round(func: Callable, ) -> Callable:
        def wrapper(x: Real, ) -> Real:
            return round(func(x), round_to, )
        return wrapper

    transform = no_transform

    if round_to is not None:
        transform = decorate_round(transform, )

    return transform
    #]




r"""
Multivariate linear least squares estimators
"""


#[

from __future__ import annotations

import numpy as _np
from numbers import Real

#]


__all__ = (
    "ordinary_least_squares",
    "destandardize_lstsq",
    "standardize",
    "destandardize",
)


def ordinary_least_squares(
    lhs: _np.ndarray,
    rhs: _np.ndarray,
) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    #[
    Mx = rhs @ rhs.T
    My = rhs @ lhs.T
    return _np.linalg.solve(Mx, My).T
    #]


def standardized_ordinary_least_squares(
    lhs: _np.ndarray,
    rhs: _np.ndarray,
    intercept: bool = True,
    dummy_obs: tuple[_np.ndarray, _np.ndarray, ] | None = None,
) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    #[
    if intercept:
        lhs_mean = None
        rhs_mean = None
    else:
        lhs_mean = 0
        rhs_mean = 0
    #
    # Standardize the data first, add dummy observations then
    lhs_standardized, lhs_mean_std, = standardize(lhs, mean=lhs_mean, )
    rhs_standardized, rhs_mean_std, = standardize(rhs, mean=rhs_mean, )
    #
    # Add dummy observations
    if dummy_obs is not None:
        lhs_dummy_obs, rhs_dummy_obs = dummy_obs
        lhs_standardized = _np.hstack((lhs_standardized, lhs_dummy_obs, ))
        rhs_standardized = _np.hstack((rhs_standardized, rhs_dummy_obs, ))
    #
    # Estimate the coefficients on standardized data
    b_standardized = ordinary_least_squares(
        lhs_standardized,
        rhs_standardized,
        intercept=False,
    )
    #
    # Destandardize the coefficients, calculate the intercept if needed
    b, c, = destandardize_lstsq(
        b_standardized,
        lhs_mean_std,
        rhs_mean_std,
        intercept=intercept,
    )
    #
    return b, c,
    #]


def destandardize_lstsq(
    bb,
    lhs_mean_std,
    rhs_mean_std,
) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    #[
    mean_lhs, std_lhs = lhs_mean_std
    mean_rhs, std_rhs = rhs_mean_std
    sigma_lhs = _np.diag(std_lhs)
    inv_sigma_rhs = _np.diag(1 / std_rhs)
    b = sigma_lhs @ bb @ inv_sigma_rhs
    c = None
    if intercept:
        c = mean_lhs - b @ mean_rhs
    return b, c,
    #]


def standardize(
    x: _np.ndarray,
    mean: _np.ndarray | Real | None = None,
) -> tuple[_np.ndarray, tuple[_np.ndarray, _np.ndarray]]:
    r"""
    """
    #[
    if mean is None:
        mean = _np.mean(x, axis=1, )
    elif isinstance(mean, Real):
        mean = _np.full((x.shape[0], ), mean, dtype=x.dtype, )
    std = _np.std(x, axis=1, mean=mean.reshape(-1, 1, ), )
    x_standardized = (x - mean.reshape(-1, 1, )) / std.reshape(-1, 1, )
    x_mean_std = (mean, std, )
    return x_standardized, x_mean_std,
    #]


def destandardize(
    x: _np.ndarray,
    mean_std : tuple[_np.ndarray, _np.ndarray],
) -> _np.ndarray:
    r"""
    """
    mean, std = mean_std
    return x * std.reshape(-1, 1, ) + mean.reshape(-1, 1, ),


def add_intercept_dummy(x: _np.ndarray, ) -> _np.ndarray:
    r"""
    """
    return _np.pad(
        x,
        ((0, 1), (0, 0), ),
        mode="constant",
        constant_values=1,
    )




"""
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )
import numpy as _np

from . import _invariants as _invariants

if TYPE_CHECKING:
    from typing import (Any, Self, Iterable, )
    from numbers import (Real, )
    from ..databoxes.main import (Databox, )
#]


class Variant:
    """
    """
    #[

    __slots__ = (
        "data",
    )

    def __init__(self, ) -> None:
        """
        """
        self.data = None

    @classmethod
    def from_databox_variant(
        klass,
        databox_v: Databox | dict,
        invariant: _invariants.Invariant,
        fallbacks: dict[str, Real] | None = None,
        overwrites: dict[str, Real] | None = None,
        clip_data_to_base_span: bool = False,
    ) -> None:
        """
        """
        def _create_nan_vector() -> _np.ndarray:
            return _np.full((invariant.num_periods, ), _np.nan, dtype=_np.float64, )
        self = klass()
        data_list = []
        for n in invariant.names:
            new_data = _create_nan_vector()
            if n in databox_v:
                new_data[:] = databox_v[n]
            data_list.append(new_data, )
        #
        self.data = _np.vstack(data_list, )
        nonbase_columns = invariant.nonbase_columns
        if clip_data_to_base_span and nonbase_columns:
            self.data[:, nonbase_columns] = _np.nan
        self._apply_fallbacks(fallbacks, invariant, )
        self._apply_overwrites(overwrites, invariant, )
        return self

    @classmethod
    def nan_data_array(
        klass,
        invariant: _invariants.Invariant,
    ) -> Self:
        """
        """
        self = klass()
        self.data = _np.full((invariant.num_names, invariant.num_periods, ), _np.nan, dtype=_np.float64, )
        return self

    def copy(self, ) -> Self:
        """
        """
        new = type(self)()
        new.data = self.data.copy()
        return new

    def update_columns_from(
        self,
        other: Self,
        columns: Iterable[int],
    ) -> None:
        """
        """
        columns = tuple(columns, )
        self.data[:, columns] = other.data[:, columns]

    def remove_periods_from_start(
        self,
        num_periods_to_remove: int,
    ) -> None:
        """
        """
        if num_periods_to_remove > 0:
            self.data = self.data[:, num_periods_to_remove:]

    def remove_periods_from_end(
        self,
        num_periods_to_remove: int,
    ) -> None:
        """
        """
        if num_periods_to_remove > 0:
            self.data = self.data[:, :-num_periods_to_remove]

    def add_periods_to_end(
        self,
        num_periods_to_add: int,
    ) -> None:
        if num_periods_to_add > 0:
            padding = ((0, 0), (0, num_periods_to_add))
            self.data = _np.pad(
                self.data,
                padding,
                mode="constant",
                constant_values=_np.nan,
            )

    def retrieve_record(
        self,
        record_id: int,
        #
        columns: slice | Iterable[int] | None = None,
    ) -> _np.ndarray:
        """
        """
        if columns is None:
            columns = ...
        return self.data[record_id, columns]

    def iter_data(self, ) -> _np.ndarray:
        """
        """
        return iter(self.data)

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        columns: slice | Iterable[int] | None = None,
    ) -> None:
        """
        """
        if columns is None:
            self.data[record_id, :] = values
        else:
            self.data[record_id, columns] = values

    def _apply_fallbacks(
        self,
        fallbacks: dict[int, Real] | None,
        invariant: _invariants.Invariant,
    ) -> None:
        """
        """
        if not fallbacks:
            return
        for record_id, name in enumerate(invariant.names, ):
            if name not in fallbacks:
                continue
            values = self.retrieve_record(record_id, )
            index_nan = _np.isnan(values)
            values[index_nan] = _np.float64(fallbacks[name])
            self.store_record(values, record_id, )

    def _apply_overwrites(
        self,
        overwrites: dict[str, Real] | None,
        invariant: _invariants.Invariant,
    ) -> None:
        """
        """
        if not overwrites:
            return
        for record_id, name in enumerate(invariant.names, ):
            if name not in overwrites:
                continue
            values = self.retrieve_record(record_id, )
            values[:] = _np.float64(overwrites[name])
            self.store_record(values, record_id, )

    def logarithmize(self, logly_indexes: tuple[int], ) -> None:
        r"""
        Logarithmize data flagged as logarithmic
        """
        if logly_indexes:
            self.data[logly_indexes, :] = _np.log(self.data[logly_indexes, :])

    def delogarithmize(self, logly_indexes: tuple[int], ) -> None:
        r"""
        Delogarithmize data flagged as logarithmic
        """
        if logly_indexes:
            self.data[logly_indexes, :] = _np.exp(self.data[logly_indexes, :])

    def rescale_data(self, factor: Real, ) -> None:
        r"""
        Rescale data by a common factor
        """
        self.data *= factor

    #]




"""
"""


#[
from __future__ import annotations

from typing import (Protocol, )
import re as _re

from ..incidences import main as _incidence
from .. import wrongdoings as _wrongdoings
#]


class LhsTransformProtocol(Protocol, ):
    _LHS_PATTERN: _re.Pattern

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
    ) -> str:
        ...


class _LhsTransform:
    """
    """
    pass


class LhsTransformNone(_LhsTransform, ):
    """
    """
    #[

    _LHS_PATTERN = _re.compile(r"(\w+)")

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        return rhs_xtring
    #]


class LhsTransformLog(_LhsTransform, ):
    """
    """
    #[

    _LHS_PATTERN = _re.compile(r"log\((\w+)\)")

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        return f"exp({rhs_xtring})"

    #]


class LhsTransformDiff(_LhsTransform, ):
    """
    """
    #[

    _LHS_PATTERN = _re.compile(r"\(\((\w+)\)-\(\1\[-1\]\)\)")

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}+({rhs_xtring})"

    #]


class LhsTransformDiffLog(_LhsTransform, ):
    """
    """
    #[

    _LHS_PATTERN = _re.compile(r"\(log\((\w+)\)-log\(\1\[-1\]\)\)")

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}*exp({rhs_xtring})"

    #]


class LhsTransformRoc(_LhsTransform, ):
    """
    """
    #[

    _LHS_PATTERN = _re.compile(r"\(\((\w+)\)/\(\1\[-1\]\)\)")

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}*({rhs_xtring})"

    #]


class LhsTransformPct(_LhsTransform, ):
    """
    """
    #[

    _LHS_PATTERN = _re.compile(r"\(100\*\((\w+)\)/\(\1\[-1\]\)-100\)")

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}*(1+({rhs_xtring})/100)"

    #]


_ALL_LHS_TRANSFORMS = (
    LhsTransformNone,
    LhsTransformLog,
    LhsTransformDiff,
    LhsTransformDiffLog,
    LhsTransformRoc,
    LhsTransformPct,
)


def recognize_transform_in_equation(
    lhs: str,
    /,
) -> tuple[LhsTransformProtocol | None, str | None]:
    """
    """
    #[
    for t in _ALL_LHS_TRANSFORMS:
        m = t._LHS_PATTERN.fullmatch(lhs)
        if m is not None:
            return t(), m.group(1)
    raise _wrongdoings.IrisPieError(f"Could not parse this LHS expression: {lhs}")
    #]




"""
"""


#[
from __future__ import annotations

from typing import (Iterable, )
import requests as _rq
import documark as _dm

from .. import dates as _dates
from ..dates import (Frequency, )
from ..series.main import (Series, )

from . import main as _databoxes
#]


_FRED_FREQ_MAP = {
    "A".casefold(): Frequency.YEARLY,
    "Q".casefold(): Frequency.QUARTERLY,
    "M".casefold(): Frequency.MONTHLY,
    "D".casefold(): Frequency.DAILY,
}


_BASE_URL = r"https://api.stlouisfed.org/fred/series"
_OBSERVATIONS_URL = _BASE_URL + r"/observations"
_API_KEY = r"951f01181da86ccb9045ce8716f82f43"
_PARAMETERS = r"?series_id={series_id}&api_key={api_key}&file_type=json"
_MISSING_VALUE = r"."


def mixin(klass, ):
    r"""
    Mix the from_fred method into the class
    """
    #[
    klass.from_fred = classmethod(from_fred, )
    return klass
    #]


@_dm.reference(category="api", )
def from_fred(
    klass,
    request: Iterable[str] | dict[str, str],
) -> _databoxes.Databox:
    r"""
................................................................................

==Download time series from FRED (St Louis Fed Database)==

This method downloads time series data from the FRED database. The data is
downloaded using the FRED API. The method requires an API key, which is provided
by the FRED website. The API key is stored in the `_API_KEY` variable in the
`_fred.py` module. The method downloads the data for the specified series IDs
and returns a `Databox` object with the downloaded series.

db = Databox.from_fred(
    request,
)

### Input arguments ###

???+ input "request"
A dictionary or list of series IDs to download from FRED. If a dictionary is
provided, the keys are used as the FRED codes and the values are used for
the names of the time series in the Databox. If list of strings is provided,
the series IDs are used as the names of the series in the `Databox` object.

### Returns ###

???+ returns "db"
A `Databox` object containing the downloaded time series data.

................................................................................
    """
    #[
    self = klass()
    if not isinstance(request, dict):
        request = _request_from_series_ids(request, )
    for name, series_id in request.items():
        self[name] = _get_series(series_id, )
    return self
    #}


def _request_from_series_ids(series_ids: Iterable[str], ):
    """
    """
    #[
    return {
        series_id.strip(): series_id.strip()
        for series_id in series_ids
    }
    #]


def _get_series(series_id: str, ):
    """
    """
    #[
    urls = _get_series_urls(series_id, )
    meta_response = _rq.get(urls["meta_url"], ).json()
    data_response = _rq.get(urls["data_url"], ).json()
    #
    if "seriess" not in meta_response or "observations" not in data_response:
        raise IriePieCritical(f"Invalid response from FRED API for {series_id}", )
    #
    freq = _get_freq_from_meta_response(meta_response, )
    iso_dates, str_values = _get_dates_and_values_from_data_response(data_response, )
    values = tuple( (float(x) if x != _MISSING_VALUE else None) for x in str_values )
    dates = _dates.periods_from_iso_strings(iso_dates, frequency=freq, )
    return Series(periods=dates, values=values, )
    #]


def _get_freq_from_meta_response(meta_response: dict, ):
    """
    """
    #[
    freq_letter = meta_response["seriess"][0]["frequency_short"].casefold()
    return _FRED_FREQ_MAP[freq_letter]
    #]


def _get_dates_and_values_from_data_response(data_response: dict, ):
    """
    """
    #[
    date_value_pairs = (
        (obs["date"], obs["value"], )
        for obs in data_response["observations"]
    )
    return zip(*date_value_pairs, )
    #]


def _get_series_urls(series_id: str, ):
    """
    """
    #[
    parameters = _PARAMETERS.format(series_id=series_id, api_key=_API_KEY, )
    meta_url = _BASE_URL + parameters
    data_url = _OBSERVATIONS_URL + parameters
    return {"meta_url": meta_url, "data_url": data_url, }
    #]




r"""
"""


#[

from __future__ import annotations

import json as _js
import documark as _dm

from ..series import Series

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any, Self

#]


def mixin(klass: type, ) -> type:
    r"""
    """
    #[
    klass.series_to_jsonable = series_to_jsonable
    klass.series_to_json_file = series_to_json_file
    klass.series_from_jsonable = classmethod(series_from_jsonable, )
    klass.series_from_json_file = classmethod(series_from_json_file, )
    return klass
    #]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Databox class
#-------------------------------------------------------------------------------


@_dm.reference(category="import_export", )
def series_to_jsonable(self, **kwargs, ) -> dict[str, Any]:
    r"""
................................................................................

==Convert all time series in the databox to a JSON-serializable dictionary==

This method converts all time series contained in the databox into a
JSON-serializable dictionary format.

    jsonable = self.series_to_jsonable(**kwargs, )


### Input arguments ###

???+ input "self"
    The databox object containing the time series to be converted.

???+ input "**kwargs"
    Additional keyword arguments to be passed to the `to_jsonable` method of each
    `Series` object. See the documentation of the `Series.to_jsonable` method in the
    `Series` class for details on the available options.


### Returns ###

???+ returns "jsonable"
    A dictionary where each key is the name of a time series in the databox and
    each value is the JSON-serializable dictionary representation of that time series.

................................................................................
    """
    #[
    return {
        key: value.to_jsonable(**kwargs, )
        for key, value in self.items()
        if isinstance(value, Series, )
    }
    #]


@_dm.reference(category="import_export", )
def series_to_json_file(
    self,
    file_name: str,
    *,
    json_dump_settings: dict | None = None,
    **kwargs,
) -> None:
    r"""
................................................................................

==Save all time series in the databox to a JSON file==

This method saves all time series contained in the databox to a JSON file.

    self.series_to_json_file(
        file_name,
        json_dump_settings=None,
        series_to_jsonable_settings=None,
    )


### Input arguments ###

???+ input "self"
    The databox object containing the time series to be saved.

???+ input "file_name"
    The name of the JSON file to which the time series will be saved.

???+ input "json_dump_settings"
    A dictionary of settings to be passed to the `json.dump` function when
    writing the JSON file. If `None`, default settings will be used alongside
    with `indent=4`.

???+ input "series_to_jsonable_settings"
    A dictionary of settings to be passed to the `series_to_jsonable` method
    of the databox. If `None`, default settings will be used.
    See [the documentation of the `Series.to_jsonable`](time_series.html#to_jsonable).

................................................................................
    """
    #[
    jsonable = self.series_to_jsonable(**kwargs, )
    json_dump_settings = json_dump_settings or {"indent": 4, }
    with open(file_name, "wt", encoding="utf-8") as f:
        _js.dump(jsonable, f, **json_dump_settings, )
    #]


@_dm.reference(
    category="constructor",
    call_name="Databox.seres_from_jsonable",
)
def series_from_jsonable(
    klass,
    jsonable: dict[str, Any],
    **kwargs,
) -> Self:
    r"""
................................................................................

==Create a databox from a JSON-serializable dictionary of time series==

This class method creates a databox from a JSON-serializable dictionary where
each key is the name of a time series and each value is the JSON-serializable
dictionary representation of that time series.

    self = Databox.series_from_jsonable(
        jsonable,
        **kwargs,
    )


### Input arguments ###

???+ input "jsonable"
    A dictionary where each key is the name of a time series and each value is
    the JSON-serializable dictionary representation of that time series.

???+ input "**kwargs"
    Additional keyword arguments to be passed to the `Series.from_jsonable`
    method. See
    [the documentation of the `Series.from_jsonable`](time_series.html#from_jsonable)
    method in the`Series` class for details on the available options.


### Returns ###

???+ returns "self"
    A Databox object containing the time series created from the JSON-serializable
    dictionary.

................................................................................
    """
    #[
    return klass({
        key: Series.from_jsonable(value, **kwargs, )
        for key, value in jsonable.items()
    })
    #]


@_dm.reference(
    category="constructor",
    call_name="Databox.seres_from_json_file",
)
def series_from_json_file(
    klass,
    file_name: str,
    json_load_settings: dict | None = None,
    **kwargs,
) -> Self:
    r"""
................................................................................

==Create a databox from a JSON file of time series==

This class method creates a databox from a JSON file containing a
JSON-serializable dictionary where each key is the name of a time series and
each value is the JSON-serializable dictionary representation of that time series.

    self = Databox.series_from_json_file(
        file_name,
        series_from_jsonable_settings=None,
        json_load_settings=None,
    )


### Input arguments ###

???+ input "file_name"
    The name of the JSON file containing the time series data.

???+ input "series_from_jsonable_settings"
    Additional dictionary with keyword arguments to be passed to the `Series.from_jsonable`
    method. See [the documentation of the `Series.from_jsonable`](time_series.html#from_jsonable).

???+ input "json_load_settings"
    A dictionary of settings to be passed to the `json.load` function when
    reading the JSON file. If `None`, default settings will be used.


### Returns ###

???+ returns "self"
    A Databox object containing the time series created from the JSON file.

................................................................................
    """
    #[
    json_load_settings = json_load_settings or {}
    with open(file_name, "rt", encoding="utf-8") as f:
        jsonable = _js.load(f, **json_load_settings, )
    return klass.series_from_jsonable(jsonable, **kwargs, )
    #]


#-------------------------------------------------------------------------------

